"""This module defines the distributions used for sampling variables."""

from __future__ import annotations

import math
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Callable, ClassVar, Type

import torch
from torch import Tensor
from torch.autograd import grad
from torch.distributions.mixture_same_family import MixtureSameFamily
from torch.distributions.multivariate_normal import MultivariateNormal

from leaspy.constants import constants
from leaspy.exceptions import LeaspyInputError
from leaspy.utils.distributions import MultinomialDistribution
from leaspy.utils.functional import NamedInputFunction
from leaspy.utils.weighted_tensor import WeightedTensor, sum_dim

__all__ = [
    "StatelessDistributionFamily",
    "StatelessDistributionFamilyFromTorchDistribution",
    "BernoulliFamily",
    "NormalFamily",
    "AbstractWeibullRightCensoredFamily",
    "WeibullRightCensoredFamily",
    "WeibullRightCensoredWithSourcesFamily",
    "SymbolicDistribution",
    "Normal",
    "Bernoulli",
    "WeibullRightCensored",
    "WeibullRightCensoredWithSources",
    # "CategoricalFamily",
    "MixtureNormalFamily",
    "MultivariateNormalFamily"
]


class StatelessDistributionFamily(ABC):
    """
    Abstract base class defining a stateless interface for distribution families.

    This class represents a family of probability distributions in a stateless manner:
    no parameters are stored in the instance. All methods operate purely via classmethods,
    using explicitly passed distribution parameters.

    Notes
    -----
    - Subclasses must define the `parameters` class variable, listing parameter names in order.
    - Each method operates solely on the passed tensors; no state or caching is assumed.

    TODO
    ----
    - Consider supporting `WeightedTensor` for distribution parameters,
      e.g., to mask latent variables like `batched_deltas` at the input level
      or directly at model parameter level (e.g., `batched_deltas_mean`).
    """

    parameters: ClassVar[tuple[str, ...]]

    @classmethod
    @abstractmethod
    def validate_parameters(cls, *params: Any) -> tuple[torch.Tensor, ...]:
        """
        Validate consistency of distribution parameters,
        returning them with out-of-place modifications if needed.
        """

    @classmethod
    def shape(cls, *params_shapes: tuple[int, ...]) -> tuple[int, ...]:
        """
        Shape of distribution samples (without any additional expansion),
        given shapes of distribution parameters.

        Parameters
        ----------
        *params_shapes : :obj:`tuple` of :obj:`int`
            The shapes of the distribution parameters, passed in the order
            defined by `cls.parameters`.

        Returns
        -------
        :obj:`tuple` of :obj:`int`
            The shape of the distribution samples.

        Raises
        ------
        :exc:`LeaspyInputError`
            If the number of provided shapes does not match the expected number of parameters.
        :exc:`NotImplementedError`
            If the distribution has no parameters, and sample shape cannot be inferred.
        """
        # We provide a default implementation which should fit for most cases
        n_params = len(params_shapes)
        if n_params != len(cls.parameters):
            raise LeaspyInputError(
                f"Expecting {len(cls.parameters)} parameters but got {n_params}"
            )
        if n_params == 0:
            raise NotImplementedError(
                "No way to infer shape of samples since no parameter"
            )
        if n_params == 1:
            return params_shapes[0]
        return torch.broadcast_shapes(*params_shapes)

    @classmethod
    @abstractmethod
    def sample(
        cls,
        *params: torch.Tensor,
        sample_shape: tuple[int, ...] = (),
    ) -> torch.Tensor:
        """
        Sample values, given distribution parameters (`sample_shape` is
        prepended to shape of distribution parameters).
        """

    @classmethod
    @abstractmethod
    def mode(cls, *params: torch.Tensor) -> torch.Tensor:
        """
        Mode of distribution (returning first value if discrete ties),
        given distribution parameters.
        """

    @classmethod
    @abstractmethod
    def mean(cls, *params: torch.Tensor) -> torch.Tensor:
        """Mean of distribution (if defined), given distribution parameters."""

    @classmethod
    @abstractmethod
    def stddev(cls, *params: torch.Tensor) -> torch.Tensor:
        """Standard-deviation of distribution (if defined), given distribution parameters."""

    @classmethod
    @abstractmethod
    def _nll(cls, x: WeightedTensor, *params: torch.Tensor) -> WeightedTensor:
        """Negative log-likelihood of value, given distribution parameters."""

    @classmethod
    @abstractmethod
    def _nll_jacobian(cls, x: WeightedTensor, *params: torch.Tensor) -> WeightedTensor:
        """Jacobian w.r.t. value of negative log-likelihood, given distribution parameters."""

    @classmethod
    def _nll_and_jacobian(
        cls,
        x: WeightedTensor,
        *params: torch.Tensor,
    ) -> tuple[WeightedTensor, WeightedTensor]:
        """Negative log-likelihood of value and its jacobian w.r.t. value, given distribution parameters."""
        # not efficient implementation by default
        return cls._nll(x, *params), cls._nll_jacobian(x, *params)

    @classmethod
    def nll(
        cls,
        x: WeightedTensor[float],
        *params: torch.Tensor,
    ) -> WeightedTensor[float]:
        """Negative log-likelihood of value, given distribution parameters."""
        return cls._nll(x, *params)

    @classmethod
    def regularization(
        cls,
        x: torch.Tensor,
        *params: torch.Tensor,
    ) -> WeightedTensor[float]:
        """Negative log-likelihood of value, given distribution parameters."""

        if isinstance (x, Tensor):
            regul = cls._nll(WeightedTensor(x), *params)
        else:
            regul = cls._nll(x, *params)
        return regul

    @classmethod
    def nll_jacobian(
        cls,
        x: WeightedTensor[float],
        *params: torch.Tensor,
    ) -> WeightedTensor[float]:
        """Jacobian w.r.t. value of negative log-likelihood, given distribution parameters."""
        return cls._nll_jacobian(x, *params)

    @classmethod
    def nll_and_jacobian(
        cls,
        x: WeightedTensor[float],
        *params: torch.Tensor,
    ) -> tuple[WeightedTensor[float], WeightedTensor[float]]:
        """Negative log-likelihood of value and its jacobian w.r.t. value, given distribution parameters."""
        return cls._nll_and_jacobian(x, *params)


class StatelessDistributionFamilyFromTorchDistribution(StatelessDistributionFamily):
    """
    Wrapper to build a `StatelessDistributionFamily` class from an existing torch distribution class.
    
    Attributes
    ----------
    dist_factory : :obj:`Callable` [...,  :class:`torch.distributions.Distribution`]
        A class variable that points to a factory function or class used to instantiate 
        the corresponding PyTorch distribution.
    """

    dist_factory: ClassVar[Callable[..., torch.distributions.Distribution]]

    @classmethod
    def validate_parameters(cls, *params: Any) -> tuple[torch.Tensor, ...]:
        """
        Validate consistency of distribution parameters, returning them with out-of-place modifications if needed.

        Parameters
        ----------
        params : Any
            The parameters to pass to the distribution factory.

        Returns
        -------
        :obj:`tuple` [ :class:`torch.Tensor`, ...] :
            The validated parameters.
        """
        distribution = cls.dist_factory(*params, validate_args=True)
        return tuple(getattr(distribution, parameter) for parameter in cls.parameters)

    @classmethod
    def sample(
        cls,
        *params: torch.Tensor,
        sample_shape: tuple[int, ...] = (),
    ) -> torch.Tensor:
        """
        Sample values, given distribution parameters (`sample_shape` is
        prepended to shape of distribution parameters).

        Parameters
        ----------
        params : :class:`torch.Tensor`
            The distribution parameters.
        sample_shape : :obj:`tuple` of :obj:`int`, optional
            Desired shape of the sample batch (prepended to the shape of a single sample).
            Defaults to an empty tuple, i.e., a single sample.

        Returns
        -------
        :class:`torch.Tensor`
            Sampled tensor
        """
        return cls.dist_factory(*params).sample(sample_shape)

    @classmethod
    def mode(cls, *params: torch.Tensor) -> torch.Tensor:
        """
        Mode of distribution (returning first value if discrete ties), given distribution parameters.

        This method should be overridden in subclasses that wrap torch distributions which
        explicitly define a mode.
  
        Parameters
        ----------
        params : :class:`torch.Tensor`
            The distribution parameters.

        Returns
        -------
        :class:`torch.Tensor` :
            The value of the distribution's mode.

        Raises
        ------
        :exc:`NotImplementedError`
            If the mode is not explicitly implemented for the specific distribution subclass.
        """
        raise NotImplementedError("Not provided in torch.Distribution interface")

    @classmethod
    def mean(cls, *params: torch.Tensor) -> torch.Tensor:
        """
        Return the mean of the distribution, if defined.

        Parameters
        ----------
        params : :class:`torch.Tensor`
            The distribution parameters.

        Returns
        -------
        :class:`torch.Tensor` :
            The value of the distribution's mean.
        """
        return cls.dist_factory(*params).mean

    @classmethod
    def stddev(cls, *params: torch.Tensor) -> torch.Tensor:
        """
        Return the standard-deviation of the distribution.

        Parameters
        ----------
        params : :class:`torch.Tensor`
            The distribution parameters.

        Returns
        -------
        :class:`torch.Tensor` :
            The value of the distribution's standard deviation.
        """
        return cls.dist_factory(*params).stddev

    @classmethod
    def _nll(cls, x: WeightedTensor, *params: torch.Tensor) -> WeightedTensor:
        """
        Compute the negative log-likelihood (NLL) for a given value under the distribution
        defined by the provided parameters.

        Parameters
        ----------
        x : :class:`~leaspy.utils.typing.WeightedTensor`
            A weighted tensor containing the values for which to compute the NLL.
            The attribute `x.value` is the tensor of values, and `x.weight` contains
            associated weights.
        params : :class:`torch.Tensor`
            Parameters of the distribution, in the order expected by `cls.dist_factory`.

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            A weighted tensor containing the NLL values (with negative log-probabilities in `.value`)
            and the same weights as the input `x`.
        """
        return WeightedTensor(-cls.dist_factory(*params).log_prob(x.value), x.weight)

    @classmethod
    def _nll_and_jacobian(
        cls,
        x: WeightedTensor,
        *params: torch.Tensor,
    ) -> tuple[WeightedTensor, WeightedTensor]:
        """
        Compute the negative log-likelihood (NLL) and its Jacobian with respect to input `x`.

        Parameters
        ----------
        x : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            A weighted tensor containing the values for which to compute the NLL and gradient.
            `x.value` is the input tensor, and `x.weight` contains the associated weights.
        params : :class:`torch.Tensor`
            Parameters of the distribution, passed to `cls.dist_factory`.

        Returns
        -------
        nll : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            The weighted negative log-likelihood.
        jacobian : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            The weighted gradient of the NLL with respect to `x.value`.
        """
        nll = cls._nll(x, *params)
        (nll_grad_value,) = grad(nll.value, (x.value,), create_graph=x.requires_grad)
        return nll, WeightedTensor(nll_grad_value, x.weight)

    @classmethod
    def _nll_jacobian(cls, x: WeightedTensor, *params: torch.Tensor) -> WeightedTensor:
        """
        Compute the Jacobian (gradient) of the negative log-likelihood (NLL) with respect to input `x`.
        
        Parameters
        ----------
        x : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            A weighted tensor for which to compute the NLL gradient.
            `x.value` is the input tensor, and `x.weight` contains the associated weights.
        params : :class:`torch.Tensor`
            Parameters of the distribution, passed to `cls.dist_factory`.

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            The weighted gradient of the NLL with respect to `x.value`.
        """
        return cls._nll_and_jacobian(x, *params)[1]


class BernoulliFamily(StatelessDistributionFamilyFromTorchDistribution):
    """
    Bernoulli family (stateless).
    
    Inherits from `StatelessDistributionFamilyFromTorchDistribution`.

    Class Attributes
    ----------------
    parameters : :obj:`tuple` of :obj:`str`
        The names of the parameters for the distribution. Here, it is `("loc",)`, where `loc` 
        represents the probability of success.
    dist_factory : :obj:`Callable`
        Reference to the torch distribution class, `torch.distributions.Bernoulli`.
    """

    parameters: ClassVar = ("loc",)
    dist_factory: ClassVar = torch.distributions.Bernoulli


class NormalFamily(StatelessDistributionFamilyFromTorchDistribution):
    """
    Normal / Gaussian family (stateless).
    
    Inherits from `StatelessDistributionFamilyFromTorchDistribution`.

    Class Attributes
    ----------------
    parameters : :obj:`tuple` of :obj:`str`
        The names of the distribution parameters: `("loc", "scale")`.
    dist_factory : :obj:`Callable`
        A reference to `torch.distributions.Normal`, the underlying PyTorch distribution.
    nll_constant_standard : :class:`torch.Tensor`
        The constant term `0.5 * log(2π)`, useful for computing the negative log-likelihood
        of the standard normal distribution.
    """

    parameters: ClassVar = ("loc", "scale")
    dist_factory: ClassVar = torch.distributions.Normal
    nll_constant_standard: ClassVar = 0.5 * torch.log(2 * torch.tensor(math.pi))

    @classmethod
    def mode(cls, loc: torch.Tensor, scale: torch.Tensor) -> torch.Tensor:
        """
        Return the mode of the distribution given the distribution's loc and scale parameters.

        Parameters
        ----------
        loc : :class:`torch.Tensor`
            The distribution loc.
        scale : :class:`torch.Tensor`
            The distribution scale.

        Returns
        -------
        :class:`torch.Tensor` :
            The value of the distribution's mode.
        """
        # `loc`, but with possible broadcasting of shape
        return torch.broadcast_tensors(loc, scale)[0]

    @classmethod
    def mean(cls, loc: torch.Tensor, scale: torch.Tensor) -> torch.Tensor:
        """
        Return the mean of the distribution, given the distribution loc and scale parameters.

        Parameters
        ----------
        loc : :class:`torch.Tensor`
            The distribution loc parameters.
        scale : :class:`torch.Tensor`
            The distribution scale parameters.

        Returns
        -------
        :class:`torch.Tensor` :
            The value of the distribution's mean.
        """
        # Hardcode method for efficiency
        # `loc`, but with possible broadcasting of shape
        return torch.broadcast_tensors(loc, scale)[0]

    @classmethod
    def stddev(cls, loc: torch.Tensor, scale: torch.Tensor) -> torch.Tensor:
        """
        Return the standard-deviation of the distribution, given loc and scale of the distribution.

        Parameters
        ----------
        loc : :class:`torch.Tensor`
            The distribution loc parameter.
        scale : :class:`torch.Tensor`
            The distribution scale parameter.

        Returns
        -------
        :class:`torch.Tensor` :
            The value of the distribution's standard deviation.
        """
        # Hardcode method for efficiency
        # `scale`, but with possible broadcasting of shape
        return torch.broadcast_tensors(loc, scale)[1]

    @classmethod
    def _nll(
        cls, x: WeightedTensor, loc: torch.Tensor, scale: torch.Tensor
    ) -> WeightedTensor:
        """
        Compute the negative log-likelihood (NLL) of a Normal distribution in a stateless manner.
      
        Parameters
        ----------
        x : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            A weighted tensor representing observed values and their associated weights.
        loc : :class:`torch.Tensor`
            The mean (location) parameter of the Normal distribution.
        scale : :class:`torch.Tensor`
            The standard deviation (scale) parameter of the Normal distribution.

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            A new `WeightedTensor` containing the computed NLL values, weighted appropriately.

        Notes
        -----
        - Broadcasting is respected between `x.value`, `loc`, and `scale`.
        - The constant `nll_constant_standard` corresponds to `0.5 * log(2π)`.
        """
        # Hardcode method for efficiency
        return WeightedTensor(
            (
                0.5 * ((x.value - loc) / scale) ** 2
                + torch.log(scale)
                + cls.nll_constant_standard
            ),
            x.weight,
        )

    @classmethod
    def _nll_jacobian(
        cls, x: WeightedTensor, loc: torch.Tensor, scale: torch.Tensor
    ) -> WeightedTensor:
        """
        Compute the Jacobian (gradient) of the negative log-likelihood (NLL) of a Normal distribution
        with respect to the observed values `x`, in a stateless and efficient manner.

        Parameters
        ----------
        x : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            A weighted tensor representing observed values and their associated weights.
        loc : :class:`torch.Tensor`
            The mean (location) parameter of the Normal distribution.
        scale : :class:`torch.Tensor`
            The standard deviation (scale) parameter of the Normal distribution.

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            A weighted tensor containing the computed gradient values, weighted accordingly.
        """
        # Hardcode method for efficiency
        return WeightedTensor((x.value - loc) / scale**2, x.weight)

    @classmethod
    def _nll_and_jacobian(
        cls,
        x: WeightedTensor,
        loc: torch.Tensor,
        scale: torch.Tensor,
    ) -> tuple[WeightedTensor, WeightedTensor]:
        """
        Compute both the negative log-likelihood (NLL) and its Jacobian (gradient) with respect to
        the observed values `x` for a Normal distribution, using an efficient hardcoded formula.
        
        Parameters
        ----------
        x : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            Weighted tensor of observed values and their weights.
        loc : :class:`torch.Tensor`
            Mean (location) parameter of the Normal distribution.
        scale : :class:`torch.Tensor`
            Standard deviation (scale) parameter of the Normal distribution.

        Returns
        -------
        tuple of two :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            - The first element is the computed NLL as a weighted tensor.
            - The second element is the Jacobian (gradient) of the NLL with respect to `x`,
            also as a weighted tensor.
        """
        # Hardcode method for efficiency
        z = (x.value - loc) / scale
        nll = 0.5 * z**2 + torch.log(scale) + cls.nll_constant_standard
        return WeightedTensor(nll, x.weight), WeightedTensor(z / scale, x.weight)

    # @classmethod
    # def sample(cls, loc, scale, *, sample_shape = ()):
    #    # Hardcode method for efficiency? (<!> broadcasting)

class MultivariateNormalFamily(StatelessDistributionFamily):
    """Multivariate Normal family with diagonal covariance (stateless)."""

    parameters: ClassVar = ("loc", "scale")  # scale = stddev for each dim
    dist_factory: ClassVar = MultivariateNormal
    nll_constant_standard: ClassVar = 0.5 * torch.log(2 * torch.tensor(math.pi))

    @classmethod
    def multi_dist_factory(cls,
                     loc: torch.Tensor,
                     scale: torch.Tensor,
                     ) -> MultivariateNormal:

        loc = torch.tensor(loc, dtype=torch.float32)
        scale = torch.diag(scale)
        scale = torch.tensor(scale, dtype=torch.float32)

        return MultivariateNormal(loc, scale)
    """
    @classmethod
    def sample(
        cls,
        *params: torch.Tensor,
        sample_shape: tuple[int, ...] = (),
    ) -> torch.Tensor:

        multi_dist = cls.multi_dist_factory(*params)
        print(multi_dist)

        return multi_dist.sample(sample_shape)
    """
    @classmethod
    def mode(cls, loc: torch.Tensor, scale: torch.Tensor) -> torch.Tensor:
        return loc  # mode = mean for Gaussian

    @classmethod
    def mean(cls, loc: torch.Tensor, scale: torch.Tensor) -> torch.Tensor:
        return loc

    @classmethod
    def stddev(cls, loc: torch.Tensor, scale: torch.Tensor) -> torch.Tensor:
        return scale

    @classmethod
    def _nll(
        cls, x: WeightedTensor, loc: torch.Tensor, scale: torch.Tensor
    ) -> WeightedTensor:

        if x.value.ndimension() != loc.ndimension():
            if loc.ndimension()==1:
                x_value_expanded = x.value.unsqueeze(1).repeat(1, loc.shape[0], 1, 1)
                #x_weight_expanded = x.weight.unsqueeze(1).repeat(1, loc.shape[0], 1, 1)
                x = WeightedTensor(x_value_expanded)
            else:
                x_value_expanded = x.value.unsqueeze(1).repeat(1, loc.shape[1], 1, 1)
                x_weight_expanded = x.weight.unsqueeze(1).repeat(1, loc.shape[1], 1, 1)
                x = WeightedTensor(x_value_expanded, x_weight_expanded)

        #print('loc.shape : ', loc.shape)
        z = (x.value - loc) / scale
        LOG_2PI = torch.log(torch.tensor(2 * torch.pi))

        nll = 0.5 * z ** 2 + torch.log(scale) + LOG_2PI
        #nll = 0.5 * torch.sum(z**2 + 2 * torch.log(scale) + torch.log(2 * torch.pi), dim=-1)
        if x.value.ndimension() != nll.ndimension():
            x_value_expanded = x.value.unsqueeze(1).repeat(1, x.shape[0], 1, 1)
            x_weight_expanded = x.weight.unsqueeze(1).repeat(1, x.shape[0], 1, 1)
            x = WeightedTensor(x_value_expanded, x_weight_expanded)

            z = (x.value - loc) / scale
            LOG_2PI = torch.log(torch.tensor(2 * torch.pi))

            nll = 0.5 * z ** 2 + torch.log(scale) + LOG_2PI

        return WeightedTensor(nll, x.weight)

    @classmethod
    def _nll_jacobian(
        cls, x: WeightedTensor, loc: torch.Tensor, scale: torch.Tensor
    ) -> WeightedTensor:
        # Gradient of NLL w.r.t loc (mean), assumes diagonal covariance
        grad = (x.value - loc) / scale**2
        return WeightedTensor(grad, x.weight)

    @classmethod
    def _nll_and_jacobian(
        cls,
        x: WeightedTensor,
        loc: torch.Tensor,
        scale: torch.Tensor,
    ) -> tuple[WeightedTensor, WeightedTensor]:
        z = (x.value - loc) / scale
        LOG_2PI = torch.log(torch.tensor(2 * torch.pi))

        nll = 0.5 * z ** 2 + torch.log(scale) + LOG_2PI
        #nll = 0.5 * torch.sum(z**2 + 2 * torch.log(scale) + torch.log(2 * torch.pi), dim=-1)
        grad = z / scale
        return WeightedTensor(nll, x.weight), WeightedTensor(grad, x.weight)


# class CategoricalFamily(StatelessDistributionFamilyFromTorchDistribution):
#    """
#    Categorical family (stateless).
#    """

#    parameters: ClassVar = ("probs",)
#    dist_factory: ClassVar = torch.distributions.Categorical

#    @classmethod
#    def extract_n_clusters(cls, probs: torch.Tensor) -> int:
#        return probs.size()[0]

#    @classmethod
#    def mixing_probabilities (cls, probs: torch.Tensor) -> torch.Tensor:
#        return torch.tensor([probs])


class MixtureNormalFamily(StatelessDistributionFamily):
    """
    A stateless mixture of univariate normal distributions.

    This class defines a mixture distribution where each component is a univariate
    normal distribution, and the mixture weights are defined by a categorical distribution.

    Parameters:
        - loc (Tensor): Mean of each normal component, one for each cluster.
        - scale (Tensor): Standard deviation of each normal component.
        - probs (Tensor): Probabilities associated with each cluster; must sum to 1.

    The mixture is modeled using `torch.distributions.MixtureSameFamily`, where:
        - The mixing distribution is a `Categorical` defined by `probs`.
        - The component distribution is `Normal(loc, scale)`.
    """

    parameters: ClassVar = ("loc", "scale", "probs")
    nll_constant_standard: ClassVar = 0.5 * torch.log(2 * torch.tensor(math.pi))

    @classmethod
    def dist_factory(
        cls,
        loc: torch.Tensor,
        scale: torch.Tensor,
        probs: torch.Tensor,
    ) -> MixtureSameFamily:
        """
        Construct a MixtureSameFamily distribution of univariate normal components.

        Parameters
        ----------
        loc : :class:`torch.Tensor`
            Mean(s) of the normal components. Shape should be broadcastable with `scale` and `probs`.
        scale : :class:`torch.Tensor`
            Standard deviation(s) of the normal components.
        probs : :class:`torch.Tensor`
            Probabilities of each mixture component. Must sum to 1 along the component axis.

        Returns
        -------
        :class: `.MixtureSameFamily`
            A mixture distribution with categorical mixing and normal components.
        """
        from torch.distributions import Categorical, Normal

        return MixtureSameFamily(
            Categorical(probs),
            Normal(loc, scale),
        )

    @classmethod
    def sample(cls, *params: torch.Tensor, sample_shape: tuple[int, ...] = ()) -> torch.Tensor:
        """
        Draw samples from the mixture of normal distributions.

        Parameters
        ----------
        *params : :class:`torch.Tensor`
            Distribution parameters in the order (loc, scale, probs). These should be
            broadcastable to define a valid MixtureSameFamily distribution.
        sample_shape : :obj:`tuple` of :obj:`int`, optional
            The desired sample shape. Defaults to an empty tuple for a single sample.

        Returns
        -------
        :class:`torch.Tensor`
            A tensor of samples drawn from the specified mixture distribution.
            The shape is `sample_shape + batch_shape`.
        """
        dist = cls.dist_factory(*params)

        return dist.sample(sample_shape)

    @classmethod
    def set_component_distribution(
        cls,
        component_distribution: torch.distributions,
        loc: torch.Tensor,
        scale: torch.Tensor,
    ) -> torch.distributions:
        """
        Ensure that the component distribution is an instance of the :class:`torch.distributions.Normal`.

        Parameters
        ----------
        component_distribution : :class:`torch.distributions.Distribution`
            The distribution object to validate. Must be an instance of `torch.distributions.Normal`.
        loc : :class:`torch.Tensor`
            Mean(s) for the normal distribution.
        scale : :class:`torch.Tensor`
            Standard deviation(s) for the normal distribution.

        Returns
        -------
        :class:`torch.distributions.Distribution`
            The newly set Normal distribution instance.

        Raises
        ------
        :exc:`ValueError`
            If `component_distribution` is not an instance of `torch.distributions.Normal`.
        """
        if not isinstance(
            cls.dist_mixture.component_distribution, torch.distributions.Normal
        ):
            raise ValueError(
                "The Component distribution need to be an "
                "instance of torch.distributions.Normal"
                "Setting the distribution to Normal"
            )
        cls.dist_mixture.component_distribution = torch.distributions.Normal(loc, scale)
        return cls.dist_mixture.component_distribution

    @classmethod
    def extract_probs(cls, *params: Any) -> torch.Tensor:
        """
        Extract the mixture probabilities from the distribution parameters.

        Parameters
        ----------
        *params : :obj:`Any`
            Distribution parameters (expected: loc, scale, probs), passed to `dist_factory`.

        Returns
        -------
        :class:`torch.Tensor`
            A 1D tensor of probabilities for each component in the mixture.
        """
        return cls.dist_factory(*params).mixture_distribution.probs

    @classmethod
    def extract_n_clusters(cls, *params: Any) -> int:
        """
        Return the number of mixture components (i.e., clusters) in the distribution.

        Parameters
        ----------
        *params : :obj:`Any`
            Distribution parameters (expected: loc, scale, probs), passed to `extract_probs`.

        Returns
        -------
        :obj:`int`
            The number of clusters/components in the mixture distribution.
        """
        return cls.extract_probs(*params).size()[0]

    @classmethod
    def extract_cluster_parameters(
        cls,
        which_cluster: int,
    ) -> tuple[Tensor, Tensor, Tensor]:
        """
        Extract the parameters (probability, mean, standard deviation) for a specific cluster.

        Parameters
        ----------
        which_cluster : :obj:`int`
            The index of the cluster to extract parameters for.

        Returns
        -------
        :obj:`tuple` of :class:`torch.Tensor`
            A tuple containing:
            - prob: Probability of the selected cluster.
            - loc: Mean of the selected normal component.
            - scale: Standard deviation of the selected normal component.
        """
        prob = torch.Tensor(cls.extract_probs)[which_cluster]
        loc = torch.Tensor(cls.mean)[which_cluster]
        scale = torch.Tensor(cls.stddev)[which_cluster]
        return prob, loc, scale

    @classmethod
    def mode(cls, *params: Any) -> torch.Tensor:
        """
        Return the mode of the mixture distribution.

        For a mixture of normal distributions, this returns the mean of the overall
        mixture distribution, which serves as a proxy for the mode.

        Parameters
        ----------
        *params : :obj:`Any`
            Distribution parameters (expected: loc, scale, probs), passed to `dist_factory`.

        Returns
        -------
        :class:`torch.Tensor`
            The mode of the mixture distribution.
        """
        return cls.dist_factory(*params).mean

    @classmethod
    def mean(cls, *params: Any) ->  torch.Tensor:
        """
        Return the mean of the mixture distribution.

        For a mixture of normal distributions, this returns the mean of the overall
        mixture distribution.

        Parameters
        ----------
        *params : :obj:`Any`
            Distribution parameters (expected: loc, scale, probs), passed to `dist_factory`.

        Returns
        -------
        :class:`torch.Tensor`
            The mean of the mixture distribution.
        """
        return cls.dist_factory(*params).mean

    @classmethod
    def stddev(cls, *params: Any) ->  torch.Tensor:
        """
        Return the standard deviation of the mixture distribution.

        For a mixture of normal distributions, this returns the standard deviation
        of the overall mixture distribution.

        Parameters
        ----------
        *params : :obj:`Any`
            Distribution parameters (expected: loc, scale, probs), passed to `dist_factory`.

        Returns
        -------
        :class:`torch.Tensor`
            The standard deviation of the mixture distribution.
        """
        return cls.dist_factory(*params).stddev

    @classmethod
    def _nll(
            cls, x: WeightedTensor, loc: torch.Tensor, scale: torch.Tensor, probs: torch.Tensor,
    ) -> WeightedTensor:
        """
        Compute the element-wise negative log-likelihood (NLL) for a mixture of normal distributions.

        This method handles two scenarios:
        - If `loc` has more than one dimension, it's assumed to represent multiple sources 
        per individual (e.g., shape [n_inds, n_clusters]).
        - If `loc` is one-dimensional, it's treated as a simple mixture (e.g., for global parameters
        like tau or xi), and broadcasting is used accordingly.

        Parameters
        ----------
        x : :class:`.WeightedTensor`
            A tensor-like object containing `value` (data tensor) and `weight` (observation weights).
        loc : :class:`torch.Tensor`
            The means of the normal components. Shape depends on the context:
            - For individual sources: [n_inds, n_clusters]
            - For shared/global parameters: [n_clusters]
        scale : :class:`torch.Tensor`
            The standard deviations of the normal components. Shape should match `loc`.
        probs : :class:`torch.Tensor`
            The mixture probabilities for each component. Not used directly in this method,
            but passed for consistency with the full distribution signature.

        Returns
        -------
        :class:`.WeightedTensor`
            The element-wise negative log-likelihood for each data point, weighted by `x.weight`.
        """

        z_list = []

        if loc.ndim > 1:  # for sources !will need modification if we mess with the sources_std as well
            n_clusters = loc.shape[1]

            for i in range(n_clusters):
                z_cluster = (x.value - loc[:, i]) / scale  # shape: [n_inds, n_sources]
                z_list.append(z_cluster)

            z = torch.stack(z_list, dim=-1)  # shape: [n_inds, n_sources, n_clusters]

        else:  # for tau and xi
            n_clusters = loc.shape[0]

            for i in range(n_clusters):
                #print(scale)
                z_cluster = ((x.value - loc[i]) / scale[i]).squeeze(1)  # shape: [n_inds]
                z_list.append(z_cluster)

            z = torch.stack(z_list, dim=1)  # shape: [n_inds,n_clusters]

        return WeightedTensor((0.5 * z ** 2
                               + torch.log(scale)
                               + cls.nll_constant_standard),x.weight,)

    @classmethod
    def _nll_jacobian(
            cls, x: WeightedTensor, loc: torch.Tensor, scale: torch.Tensor, probs:torch.Tensor,
    ) -> WeightedTensor:
        """
        Compute the Jacobian (gradient w.r.t. `loc`) of the negative log-likelihood 
        for a mixture of normal distributions.
        
        Parameters
        ----------
        x : :class:`.WeightedTensor`
            A tensor-like object with `.value` as the observed data and `.weight` as sample weights.
        loc : :class:`torch.Tensor`
            The means of the normal components. Shape can be:
            - [n_inds, n_clusters] for individual sources
            - [n_clusters] for shared parameters
        scale : :class:`torch.Tensor`
            The standard deviations of the components.
        probs : :class:`torch.Tensor`
            Mixture probabilities. Not used directly here, but included for signature consistency.

        Returns
        -------
        :class:`WeightedTensor`
            A tensor of Jacobian values for each sample and cluster, with shape:
            - [n_inds, n_sources, n_clusters] or [n_inds, n_clusters], depending on input shape.
            The `.weight` is preserved from the input `x`.
        """

        z_list = []

        if loc.ndim > 1:  # for sources !will need modification if we mess with the sources_std as well
            n_clusters = loc.shape[1]

            for i in range(n_clusters):
                z_cluster = (x.value - loc[:, i]) / scale ** 2 # shape: [n_inds, n_sources]
                z_list.append(z_cluster)

            z = torch.stack(z_list, dim=-1)  # shape: [n_inds, n_sources, n_clusters]

        else:  # for tau and xi
            n_clusters = loc.shape[0]

            for i in range(n_clusters):
                z_cluster = ((x.value - loc[i]) / scale[i] ** 2).squeeze(1)  # shape: [n_inds]
                z_list.append(z_cluster)

            z = torch.stack(z_list, dim=1)  # shape: [n_inds,n_clusters]

        return WeightedTensor(z, x.weight)

    @classmethod
    def _nll_and_jacobian(
            cls,
            x: WeightedTensor,
            loc: torch.Tensor,
            scale: torch.Tensor,
            probs: torch.Tensor,
    ) -> tuple[WeightedTensor, WeightedTensor]:
        """
        Compute both the negative log-likelihood (NLL) and its Jacobian w.r.t. `loc`
        for a mixture of normal distributions.

        Parameters
        ----------
        x : :class:`WeightedTensor`
            A tensor-like object containing observed values (`value`) and observation weights (`weight`).
        loc : :class:`torch.Tensor`
            Means of the normal components. Shape:
            - [n_inds, n_clusters] for individual-specific sources.
            - [n_clusters] for global/shared parameters.
        scale : :class:`torch.Tensor`
            Standard deviations of the normal components.
        probs : :class:`torch.Tensor`
            Mixture probabilities (not used in this method, but kept for signature consistency).

        Returns
        -------
        :obj:`tuple` of :class:`WeightedTensor`
            - nll : :class:`WeightedTensor`
                Element-wise negative log-likelihood values.
            - jacobian : :class:`WeightedTensor`
                The Jacobian (gradient) of the NLL with respect to `loc`. Equal to `z / scale`.
        """
        z_list = []

        if loc.ndim > 1:  # for sources !will need modification if we mess with the sources_std as well
            n_clusters = loc.shape[1]

            for i in range(n_clusters):
                z_cluster = (x.value - loc[:, i]) / scale  # shape: [n_inds, n_sources]
                z_list.append(z_cluster)

            z = torch.stack(z_list, dim=-1)  # shape: [n_inds, n_sources, n_clusters]

        else:  # for tau and xi
            n_clusters = loc.shape[0]

            for i in range(n_clusters):
                z_cluster = ((x.value - loc[i]) / scale[i]).squeeze(1)  # shape: [n_inds]
                z_list.append(z_cluster)

            z = torch.stack(z_list, dim=1)  # shape: [n_inds,n_clusters]

        nll = 0.5 * z ** 2 + torch.log(scale) + cls.nll_constant_standard
        return WeightedTensor(nll, x.weight), WeightedTensor(z / scale, x.weight)


class AbstractWeibullRightCensoredFamily(StatelessDistributionFamily):
    dist_weibull: ClassVar = torch.distributions.weibull.Weibull
    precision: float = 0.0001

    @classmethod
    def validate_parameters(cls, *params: Any) -> tuple[torch.Tensor, ...]:
        """Validate consistency of distribution parameters, returning them with out-of-place modifications if needed.

        Parameters
        ----------
        params : Any
            The parameters to pass to the distribution factory.

        Returns
        -------
        :obj:`tuple` [ :class:`torch.Tensor`, ...] :
            The validated parameters.
        """
        raise NotImplementedError("Validate parameters not implemented")

    @classmethod
    def sample(
        cls,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
        sample_shape: tuple[int, ...] = (),
    ) -> torch.Tensor:
        """
        Sample values from a Weibull distribution.

        Parameters
        ----------
        nu : :class:`torch.Tensor`
        rho : :class:`torch.Tensor`
        xi : :class:`torch.Tensor`
        tau : :class:`torch.Tensor`
        sample_shape : :obj:`tuple` of :obj:`int`, optional
            Shape of the samples to draw

        Returns
        -------
        :class:`torch.Tensor`
            Samples drawn from the transformed Weibull distribution, shaped according to
            `sample_shape` combined with the distribution parameter shapes.
        """
        return cls.dist_weibull(nu * torch.exp(-xi), rho).sample(sample_shape) + tau

    @classmethod
    def mode(cls, *params: torch.Tensor) -> torch.Tensor:
        """Return the mode of the distribution (returning first value if discrete ties).

        Parameters
        ----------
        params : :class:`torch.Tensor`
            The distribution parameters.

        Returns
        -------
        :class:`torch.Tensor` :
            The value of the distribution's mode.
        """
        raise NotImplementedError("Mode not implemented")

    @classmethod
    def mean(
        cls,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
    ) -> torch.Tensor:
        """Return the mean of the distribution, if defined.

        Parameters
        ----------
        nu : :class:`torch.Tensor`

        rho : :class:`torch.Tensor`

        xi : :class:`torch.Tensor`

        tau : :class:`torch.Tensor`

        Returns
        -------
        :class:`torch.Tensor` :
            The value of the distribution's mean.
        """
        return cls.dist_weibull(cls._extract_reparametrized_nu(nu, xi), rho).mean + tau

    @staticmethod
    @abstractmethod
    def _extract_reparametrized_nu(nu: torch.Tensor, xi: torch.Tensor) -> torch.Tensor:
        """Reparametrization of nu using individual parameter xi."""

    @classmethod
    def stddev(
        cls,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
    ) -> torch.Tensor:
        """Return the standard-deviation of the distribution, given distribution parameters.

        Parameters
        ----------
        nu : :class:`torch.Tensor`

        rho : :class:`torch.Tensor`

        xi : :class:`torch.Tensor`

        tau : :class:`torch.Tensor`

        Returns
        -------
        :class:`torch.Tensor` :
            The value of the distribution's standard deviation.
        """
        return cls.dist_weibull(
            cls._extract_reparametrized_nu(nu, rho, xi, tau), rho
        ).stddev

    @classmethod
    def _extract_reparametrized_parameters(
        cls,
        x: WeightedTensor,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
        *params: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Extract reparametrized parameters for the distribution given inputs.

        Parameters
        ----------
        x : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            The observed data values with associated weights.
        nu : :class:`torch.Tensor`
        rho : :class:`torch.Tensor`
        xi : :class:`torch.Tensor`
        tau : :class:`torch.Tensor`
        params :class:`torch.Tensor`
            Additional parameters for reparametrization if needed.

        Returns
        -------
        event_reparametrized_time : :class:`torch.Tensor`
            The reparametrized event times extracted from `x` and `tau`.
        weight : :class:`torch.Tensor`
            The weights associated with the original data `x`.
        nu_reparametrized : :class:`torch.Tensor`
            The reparametrized `nu` parameter combining `nu`, `rho`, `xi`, `tau`, and extra params.
        """
        # Construct reparametrized variables
        event_reparametrized_time = cls._extract_reparametrized_event(x.value, tau)
        nu_reparametrized = cls._extract_reparametrized_nu(nu, rho, xi, tau, *params)
        return event_reparametrized_time, x.weight, nu_reparametrized

    @classmethod
    def compute_log_likelihood_hazard(
        cls,
        x: WeightedTensor,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
        *params: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute the log hazard component of the likelihood for given data and parameters.

        Parameters
        ----------
        x : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            The observed data values with associated weights.
        nu : :class:`torch.Tensor`
        rho : :class:`torch.Tensor`
        xi : :class:`torch.Tensor`
        tau : :class:`torch.Tensor`
        params :class:`torch.Tensor`
            Additional parameters for reparametrization if needed.

        Returns
        -------
        :class:`torch.Tensor`
            The log hazard values for each observation, zeroed out for censored data.
        """
        event_reparametrized_time, event_bool, nu_reparametrized = (
            cls._extract_reparametrized_parameters(x, nu, rho, xi, tau, *params)
        )
        # Hazard neg log-likelihood only for patient with event not censored
        hazard = torch.where(
            event_reparametrized_time > 0,
            (rho / nu_reparametrized)
            * ((event_reparametrized_time / nu_reparametrized) ** (rho - 1.0)),
            -constants.INFINITY,
        )
        log_hazard = torch.where(hazard > 0, torch.log(hazard), hazard)
        log_hazard = torch.where(event_bool != 0, log_hazard, 0.0)
        return log_hazard

    @classmethod
    def compute_hazard(
        cls,
        x: WeightedTensor,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
        *params: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute the hazard function values for given observations and parameters.

        Parameters
        ----------
        x : :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            The observed data values with associated weights.
        nu : :class:`torch.Tensor`
        rho : :class:`torch.Tensor`
        xi : :class:`torch.Tensor`
        tau : :class:`torch.Tensor`
        params :class:`torch.Tensor`
            Additional parameters for reparametrization if needed.

        Returns
        -------
        :class:`torch.Tensor`
            Hazard values computed for each observation.
        """
        event_reparametrized_time, _, nu_reparametrized = (
            cls._extract_reparametrized_parameters(x, nu, rho, xi, tau, *params)
        )
        # Hazard neg log-likelihood only for patient with event not censored
        hazard = torch.where(
            event_reparametrized_time > 0,
            (rho / nu_reparametrized)
            * ((event_reparametrized_time / nu_reparametrized) ** (rho - 1.0)),
            0.0,
        )
        return hazard

    @classmethod
    def compute_log_survival(
        cls,
        x: torch.Tensor,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
        *params: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute the log survival function for the Weibull distribution
        given observations and parameters.
        
        Parameters
        ----------
        x : :class:`torch.Tensor`
        nu : :class:`torch.Tensor`
        rho : :class:`torch.Tensor`
        xi : :class:`torch.Tensor`
        tau ::class:`torch.Tensor`
        params : :class:`torch.Tensor`
            Additional optional parameters used in reparametrization.

        Returns
        -------
        :class:`torch.Tensor`
            Log survival values for each observation.
        """
        event_reparametrized_time, _, nu_reparametrized = (
            cls._extract_reparametrized_parameters(x, nu, rho, xi, tau, *params)
        )
        return -(
            (torch.clamp(event_reparametrized_time, min=0.0) / nu_reparametrized) ** rho
        )

    @classmethod
    def compute_predictions(
        cls,
        x: torch.Tensor,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
        *params: torch.Tensor,
    ) -> torch.Tensor:
        """
        Compute predicted survival or cumulative incidence probabilities for time-to-event data
        using a reparametrized Weibull model.
        
        Parameters
        ----------
        x : :class:`torch.Tensor`
        nu : :class:`torch.Tensor`
        rho : :class:`torch.Tensor`
        xi : :class:`torch.Tensor`
        tau ::class:`torch.Tensor`
        params : :class:`torch.Tensor`
            Additional optional parameters used in reparametrization.

        Returns
        -------
        :class:`torch.Tensor`
            Predicted survival probabilities or cumulative incidence values (depending on event type count),
            normalized by baseline survival at the last visit.
        """
        nb_events = nu.shape[0]

        # consider that the first time to predict was the last visit and is a reference point
        # and compute the survival S0
        init_log_survival = cls.compute_log_survival(
            WeightedTensor(x.value.min()), nu, rho, xi, tau, *params
        )
        init_survival = torch.exp(init_log_survival.sum(axis=1).expand(nb_events, -1).T)

        if nb_events == 1:
            # when there is only one event, we are interested in the corrected survival S/S0 (Rizopoulos, 2012, p173)
            return (
                torch.exp(cls.compute_log_survival(x, nu, rho, xi, tau, *params))
                / init_survival
            )
        else:
            # When there are multiple event we are interested in the cumulative incidence corrected: CIF/S0
            # see (Andrinopoulou, 2015)
            # Compute for all the possible points till the max
            time = WeightedTensor(
                torch.arange(
                    float(tau),
                    max(float(tau) + cls.precision, x.value.max()),
                    cls.precision,
                    dtype=float,
                )
                .expand(nb_events, -1)
                .T
            )
            log_survival = cls.compute_log_survival(time, nu, rho, xi, tau, *params)
            hazard = cls.compute_hazard(time, nu, rho, xi, tau, *params)
            total_survival = torch.exp(log_survival.sum(axis=1).expand(nb_events, -1).T)
            incidence = total_survival * hazard

            def get_cum_incidence(t, time_ix, incidence_ix):
                # t<tau then the result is 0 as survival is defined to be 1
                index = (time_ix * (time_ix <= t)).argmax() + 1
                return torch.trapezoid(incidence_ix[:index], time_ix[:index])

            list_to_cat = [
                torch.clone(x.value)
                .apply_(lambda t: get_cum_incidence(t, time.value.T[i], incidence.T[i]))
                .T
                for i in range(nb_events)
            ]
            res = torch.cat(list_to_cat).T
            return res / init_survival

    @staticmethod
    def _extract_reparametrized_event(
        event_time: torch.Tensor, tau: torch.Tensor
    ) -> torch.Tensor:
        """
        Compute reparametrized event time by shifting the original event time by `tau`.

        Parameters
        ----------
        event_time : :class:`torch.Tensor`
            Original event or censoring times for each individual.
        tau : :class:`torch.Tensor`
            Time shift (reference point), typically representing baseline time (e.g., last visit).

        Returns
        -------
        :class:`torch.Tensor`
            Reparametrized event times: event_time - tau.
        """
        return event_time - tau

    @staticmethod
    @abstractmethod
    def _extract_reparametrized_nu(
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
        *params: torch.Tensor,
    ) -> torch.Tensor:
        """Reparametrization of nu using individual parameter xi"""

    @classmethod
    def _nll(
        cls,
        x: torch.Tensor,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
        *params: torch.Tensor,
    ) -> WeightedTensor:
        """
        Compute the negative log-likelihood.

        Parameters
        ----------
        x : :class:`torch.Tensor`
        nu : :class:`torch.Tensor`
        rho : :class:`torch.Tensor`
        xi : :class:`torch.Tensor`
        tau ::class:`torch.Tensor`
        params : :class:`torch.Tensor`
            Additional optional parameters used in reparametrization.

        Returns
        -------
        :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            The computed negative log-likelihood
        """
        log_survival = cls.compute_log_survival(x, nu, rho, xi, tau, *params)
        log_hazard = cls.compute_log_likelihood_hazard(x, nu, rho, xi, tau, *params)

        return WeightedTensor(-1 * (log_survival + log_hazard))

    @classmethod
    def _nll_and_jacobian(
        cls,
        x: WeightedTensor,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
    ) -> tuple[WeightedTensor, WeightedTensor]:
        """
        Compute both the negative log-likelihood (NLL) and its Jacobian (gradient) with respect to
        the observed values `x` for a Normal distribution, using an efficient hardcoded formula.

        Parameters
        ----------
        x : :class:`torch.Tensor`
        nu : :class:`torch.Tensor`
        rho : :class:`torch.Tensor`
        xi : :class:`torch.Tensor`
        tau ::class:`torch.Tensor`
        params : :class:`torch.Tensor`
            Additional optional parameters used in reparametrization.

        Returns
        -------
        tuple of two :class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`
            - The first element is the computed NLL as a weighted tensor.
            - The second element is the Jacobian (gradient) of the NLL with respect to `x`,
        """
        return cls._nll(x, nu, rho, xi, tau), cls._nll_jacobian(x, nu, rho, xi, tau)

    @classmethod
    def _nll_jacobian(
        cls,
        x: WeightedTensor,
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
    ) -> torch.Tensor:
        pass  # WIP
        # Get inputs
        xi_format = xi[:, 0]
        event_rep_time, event_bool, nu_rep = self._extract_reparametrized_parameters(
            x, nu, rho, xi, tau
        )

        # Survival
        log_survival = cls.compute_log_survival(x, nu, rho, xi, tau)

        # Gradients
        grad_xi = rho * log_survival - event_bool * rho
        grad_tau = (rho / nu_rep * torch.exp(xi_format)) * (
            (event_rep_time / nu_rep) ** (rho - 1.0)
        ) + event_bool * (rho - 1) / event_rep_time

        # Normalise as compute on normalised variables
        to_cat = [
            grad_xi,
            grad_tau,
        ]

        grads = torch.cat(to_cat, dim=-1).squeeze(0)

        return grads


class WeibullRightCensoredFamily(AbstractWeibullRightCensoredFamily):
    parameters: ClassVar = ("nu", "rho", "xi", "tau")

    @staticmethod
    def _extract_reparametrized_nu(
        nu: torch.Tensor,
        rho: torch.Tensor,
        xi: torch.Tensor,
        tau: torch.Tensor,
        *params: torch.Tensor,
    ) -> torch.Tensor:
        """
        Extract reparametrized nu parameter.

        Parameters
        ----------
        nu : :class:`torch.Tensor`
        rho : :class:`torch.Tensor`
        xi : :class:`torch.Tensor`
        tau : :class:`torch.Tensor`
        params :class:`torch.Tensor`
            Additional parameters for reparametrization if needed.

        Returns
        -------
        :class:`torch.Tensor`
            value of reparametrized nu
        """
        return torch.exp(-xi) * nu


class WeibullRightCensoredWithSourcesFamily(AbstractWeibullRightCensoredFamily):
    parameters: ClassVar = ("nu", "rho", "xi", "tau", "survival_shifts")

    @staticmethod
    def _extract_reparametrized_nu(nu, rho, xi, tau, survival_shifts):
        return nu * torch.exp(-(xi + (1 / rho) * (survival_shifts)))


@dataclass(frozen=True)
class SymbolicDistribution:
    """Class providing symbolic methods for distribution families."""

    parameters_names: tuple[str, ...]
    dist_family: Type[StatelessDistributionFamily]

    # to hold automatic methods declared in `__post_init__`
    validate_parameters: Callable[..., tuple[torch.Tensor, ...]] = field(
        init=False, repr=False, compare=False
    )
    """Function of named distribution parameters, to validate these parameters."""

    shape: Callable[..., tuple[int, ...]] = field(init=False, repr=False, compare=False)
    """Function of named shapes of distribution parameters, to get shape of distribution samples."""

    mode: Callable[..., torch.Tensor] = field(init=False, repr=False, compare=False)
    """Function of named distribution parameters, to get mode of distribution."""

    mean: Callable[..., torch.Tensor] = field(init=False, repr=False, compare=False)
    """Function of named distribution parameters, to get mean of distribution."""

    stddev: Callable[..., torch.Tensor] = field(init=False, repr=False, compare=False)
    """Function of named distribution parameters, to get std-deviation of distribution."""

    def __post_init__(self):
        if len(self.parameters_names) != len(self.dist_family.parameters):
            raise ValueError(
                f"You provided {len(self.parameters_names)} names for {self.dist_family} parameters, "
                f"while expecting {len(self.dist_family.parameters)}: {self.dist_family.parameters}"
            )
        for bypass_method in {"validate_parameters", "shape", "mode", "mean", "stddev"}:
            object.__setattr__(self, bypass_method, self.get_func(bypass_method))

    def get_func(self, func: str, *extra_args_names: str, **kws):
        """
        Retrieve a function (e.g., 'sample', 'mode', 'mean') from the associated stateless
        distribution family, wrapped as a :class:`~leaspy.utils.functional._named_input_function.NamedInputFunction`.

        Parameters
        ----------
        func : :obj:`str`
            Name of the method to retrieve from the stateless distribution family.
        *extra_args_names : :obj:`str`
            Additional parameter names (e.g., 'sample_shape') to include before standard parameters.
        **kws : :obj:`dict`
            Optional keyword arguments passed to the :class:`~leaspy.utils.functional._named_input_function.NamedInputFunction`.

        Returns
        -------
        :class:`~leaspy.utils.functional._named_input_function.NamedInputFunction`
            A callable wrapper over the distribution method with named inputs.
        """
        return NamedInputFunction(
            getattr(self.dist_family, func),
            parameters=extra_args_names + self.parameters_names,
            kws=kws or None,
        )

    def get_func_sample(
        self, sample_shape: tuple[int, ...] = ()
    ) -> NamedInputFunction[torch.Tensor]:
        """Factory of symbolic sampling function.

        Parameters
        ----------
        sample_shape : :obj:`tuple` of :obj:`int`, optional
            The shape of the sample.
            Default=().

        Returns
        -------
        :class:`~leaspy.utils.functional.NamedInputFunction` :
            The sample function.
        """
        return self.get_func("sample", sample_shape=sample_shape)

    def get_func_sample_multivariate(
        self, sample_shape: tuple[int, ...] = ()
    ) -> NamedInputFunction[torch.Tensor]:
        """
        Factory of symbolic sampling function.

        Parameters
        ----------
        sample_shape : tuple of int, optional
            The shape of the sample.
            Default=().

        Returns
        -------
        NamedInputFunction :
            The sample function.
        """
        return self.get_func("sample", sample_shape=torch.Size([sample_shape]))

    def get_func_regularization(
        self, value_name: str
    ) -> NamedInputFunction[WeightedTensor[float]]:
        """
        Factory method to return a symbolic function computing the negative log-likelihood
        (used for regularization) from a given value.

        Parameters
        ----------
        value_name : :obj:`str`

        Returns
        -------
        :class:`~leaspy.utils.functional._named_input_function.NamedInputFunction`[:class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`]
            The named input function to use to compute negative log likelihood.
        """
        return self.get_func("regularization", value_name)

    def get_func_nll(
        self, value_name: str
    ) -> NamedInputFunction[WeightedTensor[float]]:
        """
        Factory method to return a symbolic function computing the negative log-likelihood
        (NLL) from a given value.
        
        Parameters
        ----------
        value_name : :obj:`str`

        Returns
        -------
        :class:`~leaspy.utils.functional._named_input_function.NamedInputFunction`[:class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`]
            The named input function to use to compute negative log likelihood.
        """
        return self.get_func("nll", value_name)

    def get_func_nll_jacobian(
        self, value_name: str
    ) -> NamedInputFunction[WeightedTensor[float]]:
        """Factory of symbolic function: state -> jacobian w.r.t. value of negative log-likelihood.

        Parameters
        ----------
        value_name : :obj:`str`

        Returns
        -------
        :class:`~leaspy.utils.functional.NamedInputFunction` :
            The named input function to use to compute negative log likelihood jacobian.
        """
        return self.get_func("nll_jacobian", value_name)

    def get_func_nll_and_jacobian(
        self, value_name: str
    ) -> NamedInputFunction[tuple[WeightedTensor[float], WeightedTensor[float]]]:
        """
        Factory of symbolic function: state -> (negative log-likelihood, its jacobian w.r.t. value).

        Parameters
        ----------
        value_name : :obj:`str`

        Returns
        -------
        :obj:`tuple`[:class:`~leaspy.utils.functional._named_input_function.NamedInputFunction`[:class:`~leaspy.utils.weighted_tensor._weighted_tensor.WeightedTensor`]] :
            The named input functions to use to compute negative log likelihood and its jacobian.
        """
        return self.get_func("nll_and_jacobian", value_name)

    @classmethod
    def bound_to(
        cls,
        dist_family: Type[StatelessDistributionFamily],
    ) -> Callable[..., SymbolicDistribution]:
        """
        Return a factory to create `SymbolicDistribution` bound to the provided distribution family.

        Parameters
        ----------
        dist_family : :class:`~leaspy.variables.distributions.StatelessDistributionFamily`
            The distribution family to use to create a SymbolicDistribution.

        Returns
        -------
        factory : :obj:`Callable`[..., :class:`~leaspy.variables.distributions.SymbolicDistribution`]
            The factory.
        """

        def factory(*parameters_names: str) -> SymbolicDistribution:
            """
            Factory of a `SymbolicDistribution`, bounded to the provided distribution family.

            Parameters
            ----------
            *parameters_names : :obj:`str`
                Names, in order, for distribution parameters.

            Returns
            -------
            :class:`~leaspy.variables.distributions.SymbolicDistribution` :
                The symbolic distribution resulting from the factory.
            """
            return SymbolicDistribution(parameters_names, dist_family)

        # Nicer runtime name and docstring for the generated factory function
        factory.__name__ = f"symbolic_{dist_family.__name__}_factory"
        factory.__qualname__ = ".".join(
            factory.__qualname__.split(".")[:-1] + [factory.__name__]
        )
        factory.__doc__ = factory.__doc__.replace(
            "the provided distribution family", f"`{dist_family.__name__}`"
        ).replace(
            "for distribution parameters",
            f"for distribution parameters: {dist_family.parameters}",
        )

        return factory


Normal = SymbolicDistribution.bound_to(NormalFamily)
# Categorical = SymbolicDistribution.bound_to(CategoricalFamily)
MixtureNormal = SymbolicDistribution.bound_to(MixtureNormalFamily)
MultivariateNormal = SymbolicDistribution.bound_to(MultivariateNormalFamily)
Bernoulli = SymbolicDistribution.bound_to(BernoulliFamily)
WeibullRightCensored = SymbolicDistribution.bound_to(WeibullRightCensoredFamily)
WeibullRightCensoredWithSources = SymbolicDistribution.bound_to(
    WeibullRightCensoredWithSourcesFamily
)

# INLINE UNIT TESTS
if __name__ == "__main__":
    print(Normal)
    print(Normal("mean", "std").validate_parameters(mean=0.0, std=1.0))

    nll = Normal("mean", "std").get_func_nll("val")

    args = dict(
        val=WeightedTensor(
            torch.tensor(
                [
                    [-1.0, 0.0, 0.0, 1.0],
                    [0.5, -0.5, -1.0, 0.0],
                ]
            ),
            weight=torch.tensor(
                [
                    [1, 0, 1, 1],
                    [1, 1, 0, 0],
                ]
            ),
        ),
        mean=torch.zeros((2, 4)),
        std=torch.ones(()),
    )

    r_nll = nll(**args)
    print("nll: ", r_nll)
    r_nll_sum_0 = nll.then(sum_dim, dim=1)(**args)
    r_nll_sum_1 = nll.then(sum_dim, dim=1)(**args)
    r_nll_sum_01 = nll.then(sum_dim, dim=(0, 1))(**args)  # MaskedTensor.wsum
    print("nll_sum_0: ", r_nll_sum_0)
    print("nll_sum_1: ", r_nll_sum_1)
    print("nll_sum_01: ", r_nll_sum_01)
    print("nll_sum_0,1: ", sum_dim(r_nll_sum_0, dim=0))
    print("nll_sum_1,0: ", sum_dim(r_nll_sum_1, dim=0))



import torch

from leaspy.exceptions import LeaspyModelInputError
from leaspy.utils.typing import DictParamsTorch

from .stateless import StatelessModel

__all__ = ["ConstantModel"]


class ConstantModel(StatelessModel):
    r"""ConstantModel` is a benchmark model that predicts constant values (no matter what the patient's ages are).

    These constant values depend on the algorithm setting and the patient's values
    provided during :term:`calibration`.

    It could predict:
        * ``last``: last value seen during calibration (even if ``NaN``).
        * ``last_known``: last non ``NaN`` value seen during :term:`calibration`.
        * ``max``: maximum (=worst) value seen during :term:`calibration`.
        * ``mean``: average of values seen during :term:`calibration`.

    .. warning::
        Depending on ``features``, the ``last_known`` / ``max`` value
        may correspond to different visits.

    .. warning::
        For a given feature, value will be ``NaN`` if and only if all
        values for this feature were ``NaN``.

    Parameters
    ----------
    name : :obj:`str`
        The model's name.

    See Also
    --------
    :class:`~leaspy.algo.personalize.constant_prediction_algo.ConstantPredictionAlgorithm`
    """

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)
        self._is_initialized = True

    @property
    def hyperparameters(self) -> DictParamsTorch:
        """Dictionary of values for model hyperparameters.

        Returns
        -------
        :class:`~leaspy.utils.typing.DictParamsTorch`
            Dictionary of hyperparameters.
        """
        return {}

    def compute_individual_trajectory(
        self,
        timepoints: torch.Tensor,
        individual_parameters: dict,
    ) -> torch.Tensor:
        """Compute the individual trajectory based on the model's features and parameters.

        Parameters
        ----------
        timepoints : :obj:`torch.Tensor`
            The time points at which to compute the trajectory.
        individual_parameters : :obj:`dict`
            Dictionary containing the individual's parameters, where keys are feature names.

        Returns
        -------
        :obj:`torch.Tensor`
            A tensor containing the computed trajectory for the individual.

        Raises
        ------
        :class:`~leaspy.exceptions.LeaspyModelInputError`
            If the model was not properly initialized or if features are not set.
        """
        if self.features is None:
            raise LeaspyModelInputError("The model was not properly initialized.")
        values = [individual_parameters[f] for f in self.features]
        return torch.tensor([[values] * len(timepoints)], dtype=torch.float32)



import math
import os
from itertools import cycle
from typing import Optional

import matplotlib as mpl
import matplotlib.backends.backend_pdf
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from matplotlib.lines import Line2D

from leaspy.exceptions import LeaspyInputError
from leaspy.models import McmcSaemCompatibleModel
from leaspy.utils.typing import DictParamsTorch

from ...data import Dataset

__all__ = ["Plotter"]


class Plotter:
    """
    Class defining some plotting tools.

    Parameters
    ----------
    output_path : :obj:`str`, (optional)
        Folder where plots will be saved.
        If None, default to current working directory.
    """

    def __init__(self, output_path: Optional[str] = None):
        # TODO : Do all the check up if the path exists, and if yes, if removing or not
        if output_path is None:
            output_path = os.getcwd()
        self.output_path = output_path

        # https://stackoverflow.com/questions/40395659/view-and-then-close-the-figure-automatically-in-matplotlib/40395799
        self._block = False
        self._show = True

    @staticmethod
    def _torch_model_values_to_numpy_postprocessed_values(
        model_values: torch.Tensor,
        *,
        model: McmcSaemCompatibleModel,
    ) -> np.ndarray:
        """
        Convert torch model values to numpy & apply them the default
        postprocessing (useful for ordinal models).
        """
        model_values_np = model_values.cpu().detach().numpy()
        # post-process the mean trajectory if needed (with default arguments)
        estimation_postprocessor = getattr(model, "postprocess_model_estimation", None)
        if estimation_postprocessor is not None:
            model_values_np = estimation_postprocessor(model_values_np)

        return model_values_np

    @classmethod
    def _compute_mean_traj_postprocessed(
        cls,
        model: McmcSaemCompatibleModel,
        timepoints: torch.Tensor,
    ) -> np.ndarray:
        mean_trajectory = model.compute_mean_traj(timepoints)
        return cls._torch_model_values_to_numpy_postprocessed_values(
            mean_trajectory, model=model
        )

    @classmethod
    def _compute_individual_tensorized_postprocessed(
        cls,
        model: McmcSaemCompatibleModel,
        timepoints: torch.Tensor,
        individual_parameters: DictParamsTorch,
        **kws,
    ) -> np.ndarray:
        model_values = model.compute_individual_tensorized(
            timepoints, individual_parameters, **kws
        )
        return cls._torch_model_values_to_numpy_postprocessed_values(
            model_values, model=model
        )

    @classmethod
    def _compute_individual_trajectory_postprocessed(
        cls,
        model: McmcSaemCompatibleModel,
        timepoints: torch.Tensor,
        individual_parameters: DictParamsTorch,
        **kws,
    ) -> np.ndarray:
        model_values = model.compute_individual_trajectory(
            timepoints, individual_parameters, **kws
        )
        return cls._torch_model_values_to_numpy_postprocessed_values(
            model_values, model=model
        )

    def plt_show(self):
        """Display the current matplotlib figure if `self._show` is True.

        This method wraps `matplotlib.pyplot.show` using the internal
        flags `_show` and `_block` to control interactive plotting.
        """
        if self._show:
            plt.show(block=self._block)

    def plot_mean_trajectory(
        self,
        model: McmcSaemCompatibleModel,
        *,
        n_pts: int = 100,
        n_std_left: int = 3,
        n_std_right: int = 6,
        **kwargs,
    ):
        """Plot the mean model trajectory.

        Parameters
        ----------
        model : :class:`McmcSaemCompatibleModel` or iterable of such
            Model(s) to compute the mean trajectory for.
        n_pts : :obj:`int`, optional
            Number of timepoints to evaluate between the start and end.
            Default=100.
        n_std_left : :obj:`int`, optional
            How many standard deviations before the mean to plot.
            Default=3.
        n_std_right : :obj:`int`, optional
            How many standard deviations after the mean to plot.
            Default=6.
        **kwargs :
            - ``color`` : iterable of colors
            - ``title`` : :obj:`str`, plot title
            - ``save_as`` : :obj:`str`, filename to save the plot

        Raises
        ------
        LeaspyInputError
            If the model(s) is/are not initialized.
        """
        labels = model.features
        fig, ax = plt.subplots(1, 1, figsize=(11, 6))
        colors = kwargs.get("color", cycle(mpl.colormaps["tab20"].colors))

        try:
            iter(model)
        except Exception:
            # Break if model is not initialized
            if not model.is_initialized:
                raise LeaspyInputError("Please initialize the model before plotting")

            # not iterable
            if getattr(model, "is_ordinal", False):
                ax.set_ylim(0, model.ordinal_infos["max_level"])
            elif "logistic" in model.name:
                ax.set_ylim(0, 1)

            mean_time = model.parameters["tau_mean"]
            std_time = max(model.parameters["tau_std"], 4)
            timepoints_np = np.linspace(
                mean_time - n_std_left * std_time,
                mean_time + n_std_right * std_time,
                n_pts,
            )
            mean_trajectory = self._compute_mean_traj_postprocessed(
                model,
                torch.tensor(timepoints_np).unsqueeze(0),
            )
            for i, color_ft in zip(range(mean_trajectory.shape[-1]), colors):
                ax.plot(
                    timepoints_np,
                    mean_trajectory.squeeze()[:, i],
                    label=labels[i],
                    linewidth=4,
                    alpha=0.9,
                    c=color_ft,
                )
            plt.legend()
        else:
            # Break if model is not initialized
            if not model[0].is_initialized:
                raise LeaspyInputError("Please initialize the model before plotting")

            # iterable
            if getattr(model[0], "is_ordinal", False):
                ax.set_ylim(0, model[0].ordinal_infos["max_level"])
            elif "logistic" in model[0].name:
                ax.set_ylim(0, 1)

            mean_time = model[0].parameters["tau_mean"]
            std_time = max(model[0].parameters["tau_std"], 4)
            timepoints_np = np.linspace(
                mean_time - n_std_left * std_time,
                mean_time + n_std_right * std_time,
                n_pts,
            )
            for j, model_j in enumerate(model):
                mean_trajectory_np = self._compute_mean_traj_postprocessed(
                    model_j,
                    torch.tensor(timepoints_np).unsqueeze(0),
                )
                for i, color_ft in zip(range(mean_trajectory_np.shape[-1]), colors):
                    ax.plot(
                        timepoints_np,
                        mean_trajectory_np[0, :, i],
                        label=labels[i],
                        linewidth=4,
                        alpha=0.5,
                        c=color_ft,
                    )
                if j == 0:
                    ax.legend()

        title = kwargs["title"] if "title" in kwargs.keys() else None
        if title is not None:
            ax.set_title(title)

        if "save_as" in kwargs.keys():
            plt.savefig(os.path.join(self.output_path, kwargs["save_as"]))

        self.plt_show()
        plt.close()

    def plot_mean_validity(
        self, model: McmcSaemCompatibleModel, results, **kwargs
    ) -> None:
        """Plot histogram of reparametrized times for all individuals.

        Parameters
        ----------
        model : :class:`McmcSaemCompatibleModel`
            Fitted model providing `tau_mean` and `tau_std`.
        results : object
            Results containing `data` and `individual_parameters`
            with keys ``xi`` and ``tau``.
        **kwargs :
            - ``save_as`` : :obj:`str`, filename to save the plot.
        """
        t0 = model.parameters["tau_mean"].numpy()
        hist = []

        for i, individual in enumerate(results.data):
            ages = individual.timepoints
            xi = results.individual_parameters["xi"][i].numpy()
            tau = results.individual_parameters["tau"][i].numpy()
            reparametrized = np.exp(xi) * (ages - tau) + t0
            hist.append(reparametrized)

        hist = np.concatenate(hist)
        plt.hist(hist)

        if "save_as" in kwargs.keys():
            plt.savefig(os.path.join(self.output_path, kwargs["save_as"]))

        self.plt_show()
        plt.close()

    def plot_patient_trajectory(
        self, model: McmcSaemCompatibleModel, results, indices, **kwargs
    ) -> None:
        """Plot observed data and reconstructed trajectory for one or more patients.

        Parameters
        ----------
        model : :class:`McmcSaemCompatibleModel`
            Model to use for computing individual trajectories.
        results : object
            Results containing data and individual parameters.
        indices : :obj:`str` or list of :obj:`int`
            Patient index/indices to plot.
        **kwargs :
            - ``ax`` : matplotlib axis to plot on.
            - ``color`` : iterable of colors.
            - ``title`` : :obj:`str`, plot title.
            - ``save_as`` : :obj:`str`, filename to save the plot.
        """
        colors = (
            kwargs["color"]
            if "color" in kwargs.keys()
            else mpl.colormaps["Dark2"](np.linspace(0, 1, model.dimension))
        )
        labels = model.features
        if "ax" in kwargs.keys():
            ax = kwargs["ax"]
        else:
            fig, ax = plt.subplots(1, 1, figsize=(8, 8))

        if getattr(model, "is_ordinal", False):
            ax.set_ylim(0, model.ordinal_infos["max_level"])
        elif "logistic" in model.name:
            ax.set_ylim(0, 1)

        if not isinstance(indices, list):
            indices = [indices]

        for idx in indices:
            indiv = results.data[idx]
            timepoints = indiv.timepoints
            observations = np.array(indiv.observations)
            t = torch.tensor(timepoints).unsqueeze(0)
            indiv_parameters = results.get_patient_individual_parameters(idx)

            trajectory_np = self._compute_individual_tensorized_postprocessed(
                model, t, indiv_parameters
            ).squeeze(0)
            for dim in range(model.dimension):
                not_nans_idx = np.array(~np.isnan(observations[:, dim]), dtype=bool)
                ax.plot(np.array(timepoints), trajectory_np[:, dim], c=colors[dim])
                ax.plot(
                    np.array(timepoints)[not_nans_idx],
                    observations[:, dim][not_nans_idx],
                    c=colors[dim],
                    linestyle="--",
                )

        if "title" in kwargs.keys():
            ax.set_title(kwargs["title"])

        custom_lines = [
            Line2D([0], [0], color=colors[i], lw=4) for i in range(model.dimension)
        ]

        ax.legend(custom_lines, labels, loc="upper right")

        if "save_as" in kwargs.keys():
            plt.savefig(os.path.join(self.output_path, kwargs["save_as"]))

        if "ax" not in kwargs.keys():
            self.plt_show()
            plt.close()

    def plot_from_individual_parameters(
        self,
        model: McmcSaemCompatibleModel,
        indiv_parameters: DictParamsTorch,
        timepoints: torch.Tensor,
        **kwargs,
    ) -> None:
        """Plot a trajectory computed from given individual parameters.

        Parameters
        ----------
        model : :class:`McmcSaemCompatibleModel`
            Model used to compute the trajectory.
        indiv_parameters : DictParamsTorch
            Individual parameter dictionary.
        timepoints : :obj:`torch.Tensor`
            Timepoints at which to compute the trajectory.
        **kwargs :
            - ``color`` : iterable of colors.
            - ``save_as`` : :obj:`str`, filename to save the plot.
        """
        # 1 individual at a time...
        colors = (
            kwargs["color"]
            if "color" in kwargs.keys()
            else mpl.colormaps["Dark2"](np.linspace(0, 1, model.dimension))
        )
        labels = model.features
        fig, ax = plt.subplots(1, 1, figsize=(11, 6))

        trajectory_np = self._compute_individual_trajectory_postprocessed(
            model, timepoints, indiv_parameters
        ).squeeze(0)
        for dim in range(model.dimension):
            ax.plot(timepoints, trajectory_np[:, dim], c=colors[dim], label=labels[dim])

        ax.legend()
        if "save_as" in kwargs.keys():
            plt.savefig(os.path.join(self.output_path, kwargs["save_as"]))

        self.plt_show()
        plt.close()

    def plot_distribution(self, results, parameter: str, cofactor=None, **kwargs):
        """Plot histogram(s) of an estimated parameter distribution.

        Parameters
        ----------
        results : object
            Results containing a `get_parameter_distribution` method.
        parameter : :obj:`str`
            Parameter name to plot.
        cofactor : optional
            Grouping variable; if given, plot separate histograms per group.
        **kwargs :
            - ``save_as`` : :obj:`str`, filename to save the plot.
        """
        fig, ax = plt.subplots(1, 1, figsize=(11, 6))

        distribution = results.get_parameter_distribution(parameter, cofactor)

        if cofactor is None:
            ax.hist(distribution)
        else:
            for k, v in distribution.items():
                ax.hist(v, label=k, alpha=0.7)
            plt.legend()

        if "save_as" in kwargs.keys():
            plt.savefig(os.path.join(self.output_path, kwargs["save_as"]))

        self.plt_show()
        plt.close()

    def plot_correlation(
        self, results, parameter_1, parameter_2, cofactor=None, **kwargs
    ):
        """Plot scatter correlation between two parameters.

        Parameters
        ----------
        results : object
            Results containing a `get_parameter_distribution` method.
        parameter_1 : :obj:`str`
            First parameter name.
        parameter_2 : :obj:`str`
            Second parameter name.
        cofactor : optional
            Grouping variable; if given, scatter different colors per group.
        **kwargs :
            - ``save_as`` : :obj:`str`, filename to save the plot.
        """
        fig, ax = plt.subplots(1, 1, figsize=(11, 6))

        d1 = results.get_parameter_distribution(parameter_1, cofactor)
        d2 = results.get_parameter_distribution(parameter_2, cofactor)

        if cofactor is None:
            ax.scatter(d1, d2)

        else:
            for possibility in d1.keys():
                ax.scatter(d1[possibility], d2[possibility], label=possibility)

        plt.legend()
        if "save_as" in kwargs.keys():
            plt.savefig(os.path.join(self.output_path, kwargs["save_as"]))

        self.plt_show()
        plt.close()

    def plot_patients_mapped_on_mean_trajectory(
        self,
        model: McmcSaemCompatibleModel,
        results,
        *,
        n_std_left: int = 2,
        n_std_right: int = 4,
        n_pts: int = 100,
    ) -> None:
        """Plot observed patient values mapped onto the mean trajectory.

        Parameters
        ----------
        model : :class:`McmcSaemCompatibleModel`
            Model used for computing mean and individual trajectories.
        results : object
            Results containing data and individual parameters.
        n_std_left : :obj:`int`, optional
            How many standard deviations before the mean to plot.
            Default=2.
        n_std_right : :obj:`int`, optional
            How many standard deviations after the mean to plot.
            Default=4.
        n_pts : :obj:`int`, optional
            Number of timepoints to evaluate. Default=100.
        """
        dataset = Dataset(results.data)

        model_values_np = self._compute_individual_tensorized_postprocessed(
            model, dataset.timepoints, results.individual_parameters
        )
        timepoints = np.linspace(
            model.parameters["tau_mean"] - n_std_left * model.parameters["tau_std"],
            model.parameters["tau_mean"] + n_std_right * model.parameters["tau_std"],
            n_pts,
        )
        timepoints = torch.tensor(timepoints).unsqueeze(0)
        xi = results.individual_parameters["xi"]
        tau = results.individual_parameters["tau"]

        reparametrized_time = (
            model.time_reparametrization(
                t=dataset.timepoints, alpha=torch.exp(xi), tau=tau
            )
            / torch.exp(model.parameters["xi_mean"])
            + model.parameters["tau_mean"]
        )

        for i in range(dataset.values.shape[-1]):
            fig, ax = plt.subplots(1, 1)
            for idx in range(min(50, len(tau))):
                ax.plot(
                    reparametrized_time[idx, 0 : dataset.n_visits_per_individual[idx]]
                    .cpu()
                    .detach()
                    .numpy(),
                    dataset.values[idx, 0 : dataset.n_visits_per_individual[idx], i]
                    .cpu()
                    .detach()
                    .numpy(),
                    "x",
                )
                ax.plot(
                    reparametrized_time[idx, 0 : dataset.n_visits_per_individual[idx]]
                    .cpu()
                    .detach()
                    .numpy(),
                    model_values_np[idx, 0 : dataset.n_visits_per_individual[idx], i],
                    alpha=0.8,
                )

            if getattr(model, "is_ordinal", False):
                ax.set_ylim(0, model.ordinal_infos["max_level"])
            elif "logistic" in model.name:
                ax.set_ylim(0, 1)

        self.plt_show()
        plt.close()

    ############## TODO : The next functions are related to the plots during the fit. Disentangle them properly

    @classmethod
    def plot_error(
        cls,
        path,
        dataset,
        model: McmcSaemCompatibleModel,
        param_ind,
        colors=None,
        labels=None,
    ):
        model_values_np = cls._compute_individual_tensorized_postprocessed(
            model, dataset.timepoints, param_ind
        )

        if colors is None:
            colors = mpl.colormaps["rainbow"](
                np.linspace(0, 1, model_values_np.shape[-1])
            )
        if labels is None:
            labels = np.arange(model_values_np.shape[-1])
            labels = [str(k) for k in labels]

        err = {"all": []}
        for i in range(dataset.values.shape[-1]):
            err[i] = []
            for idx in range(model_values_np.shape[0]):
                err[i].extend(
                    dataset.values[idx, 0 : dataset.n_visits_per_individual[idx], i]
                    .cpu()
                    .detach()
                    .numpy()
                    - model_values_np[idx, 0 : dataset.n_visits_per_individual[idx], i]
                )

            err["all"].extend(err[i])
            err[i] = np.array(err[i])
        err["all"] = np.array(err["all"])
        pdf = matplotlib.backends.backend_pdf.PdfPages(path)
        for i in range(dataset.values.shape[-1]):
            fig, ax = plt.subplots(1, 1)
            # sns.distplot(err[i], color='blue')
            plt.title(
                labels[i]
                + " sqrt mean square error: "
                + str(np.sqrt(np.mean(err[i] ** 2)))
            )
            pdf.savefig(fig)
            plt.close()
        fig, ax = plt.subplots(1, 1)
        # sns.distplot(err['all'], color='blue')
        plt.title(
            "global sqrt mean square error: " + str(np.sqrt(np.mean(err["all"] ** 2)))
        )
        pdf.savefig(fig)
        plt.close()
        pdf.close()

    @classmethod
    def plot_patient_reconstructions(
        cls,
        path: str,
        dataset: Dataset,
        model: McmcSaemCompatibleModel,
        param_ind: DictParamsTorch,
        *,
        max_patient_number: int = 5,
        attribute_type=None,
    ):
        if isinstance(max_patient_number, int):
            max_patient_number = min(max_patient_number, dataset.n_individuals)
            patients_list = range(max_patient_number)
            n_pats = max_patient_number
        else:
            # list of ints (not the ID but the indices of wanted patients [0, 1, 2, 3...])
            patients_list = max_patient_number
            n_pats = len(patients_list)

        colors = mpl.colormaps["Dark2"](np.linspace(0, 1, n_pats + 2))

        fig, ax = plt.subplots(1, 1)

        model_values_np = cls._compute_individual_tensorized_postprocessed(
            model,
            dataset.timepoints,
            param_ind,
            attribute_type=attribute_type,
        )
        for i in patients_list:
            times_pat = dataset.get_times_patient(i).cpu().detach().numpy()
            true_values_pat = dataset.get_values_patient(i).cpu().detach().numpy()
            model_values_pat = model_values_np[
                i, 0 : dataset.n_visits_per_individual[i], :
            ]

            ax.plot(times_pat, model_values_pat, c=colors[i])
            ax.plot(times_pat, true_values_pat, c=colors[i], linestyle="--", marker="o")

        # Plot the mean also
        # min_time, max_time = torch.min(dataset.timepoints[dataset.timepoints>0.0]), torch.max(dataset.timepoints)

        min_time, max_time = np.percentile(
            dataset.timepoints[dataset.timepoints > 0.0].cpu().detach().numpy(),
            [10, 90],
        )
        timepoints_np = np.linspace(min_time, max_time, 100)
        model_values_np = cls._compute_mean_traj_postprocessed(
            model, torch.tensor(timepoints_np).unsqueeze(0)
        )
        for ft_k in range(model.dimension):
            ax.plot(
                timepoints_np,
                model_values_np[0, :, ft_k],
                c="black",
                linewidth=3,
                alpha=0.3,
            )
        plt.savefig(path)
        plt.close()

        return ax

    @staticmethod
    def plot_param_ind(path, param_ind):
        # <!> param_ind is expected to be iterable of values not the usual dictionary

        pdf = matplotlib.backends.backend_pdf.PdfPages(path)
        fig, ax = plt.subplots(1, 1)
        if len(param_ind) == 2:
            # no sources
            xi, tau = param_ind
            sources = torch.zeros((0, 0))
        else:
            # with sources
            xi, tau, sources = param_ind
        ax.plot(
            xi.squeeze(1).cpu().detach().numpy(),
            tau.squeeze(1).cpu().detach().numpy(),
            "x",
        )
        plt.xlabel("xi")
        plt.ylabel("tau")
        pdf.savefig(fig)
        plt.close()

        nb_sources = sources.shape[1]

        for i in range(nb_sources):
            fig, ax = plt.subplots(1, 1)
            ax.plot(sources[:, i].cpu().detach().numpy(), "x")
            plt.title("sources " + str(i))
            pdf.savefig(fig)
            plt.close()
        pdf.close()

    ## TODO : Refaire avec le path qui est fourni en haut!
    @staticmethod
    def plot_convergence_model_parameters(
        path, path_saveplot_1, path_saveplot_2, model
    ):
        # TODO? add legends (color <-> feature, esp. for g/v0/noise_std/deltas)
        # TODO? add loss (log-likelihood) or some information criteria AIC/BIC

        # figure dimensions
        width = 10
        height_per_row = 3.5

        # don't keep the sources parameters (fixed mean = 0 & std = 1 by design)
        skip_sources = True

        # Make the plot 1

        to_skip_1 = ["betas"] + ["sources_mean", "sources_std"] * int(skip_sources)
        if getattr(model, "is_ordinal", False):
            to_skip_1.append("deltas")
        params_to_plot_1 = [p for p in model.parameters.keys() if p not in to_skip_1]

        n_plots_1 = len(params_to_plot_1)
        n_rows_1 = math.ceil(n_plots_1 / 2)
        _, ax = plt.subplots(n_rows_1, 2, figsize=(width, n_rows_1 * height_per_row))

        for i, key in enumerate(params_to_plot_1):
            import_path = os.path.join(path, key + ".csv")
            df_convergence = pd.read_csv(import_path, index_col=0, header=None)
            df_convergence.index.rename("iter", inplace=True)

            x_position = i // 2
            y_position = i % 2
            # ax[x_position][y_position].plot(df_convergence.index.values, df_convergence.values)
            df_convergence.plot(ax=ax[x_position][y_position], legend=False)
            ax[x_position][y_position].set_title(key)

        plt.tight_layout()
        plt.savefig(path_saveplot_1)
        plt.close()

        # Make the plot 2

        reals_pop_name = model.get_population_variable_names()
        reals_ind_name = model.get_individual_variable_names()

        additional_plots = 1  # for noise_std / log-likelihood depending on noise_model

        if skip_sources and "sources" in reals_ind_name:
            additional_plots -= 1

        n_plots_2 = len(reals_pop_name) + len(reals_ind_name) + additional_plots
        _, ax = plt.subplots(n_plots_2, 1, figsize=(width, n_plots_2 * height_per_row))

        # nonposy is deprecated since Matplotlib 3.3
        mpl_version = mpl.__version__.split(".")
        if int(mpl_version[0]) < 3 or (
            (int(mpl_version[0]) == 3) and (int(mpl_version[1]) < 3)
        ):
            yscale_kw = dict(nonposy="clip")
        else:  # >= 3.3
            yscale_kw = dict(nonpositive="clip")

        # Goodness-of-fit monitoring
        y_position = 0

        ## TODO: improve this, we could want to monitor both noise-std and log-likelihood in practice...
        # goodness_of_fit_path = os.path.join(path, 'noise_std.csv')
        # if os.path.exists(goodness_of_fit_path):
        #    goodness_of_fit_title = 'noise_std'
        # else:
        #    # LL for other models
        #    goodness_of_fit_path = os.path.join(path, 'log-likelihood.csv')
        #    goodness_of_fit_title = 'log-likelihood'
        #
        # df_convergence = pd.read_csv(goodness_of_fit_path, index_col=0, header=None)
        # df_convergence.index.rename("iter", inplace=True)
        # df_convergence.plot(ax=ax[y_position], legend=False)
        # ax[y_position].set_title(goodness_of_fit_title)
        # ax[y_position].set_yscale("log", **yscale_kw)

        for i, key in enumerate(reals_pop_name):
            y_position += 1
            ax[y_position].set_title(key)
            if key == "deltas" and getattr(model, "is_ordinal", False):
                for dim in range(model.dimension):
                    import_path = os.path.join(path, key + "_" + str(dim) + ".csv")
                    df_convergence = pd.read_csv(import_path, index_col=0, header=None)
                    df_convergence.index.rename("iter", inplace=True)
                    df_convergence.plot(ax=ax[y_position], legend=False)
            elif key != "betas":
                import_path = os.path.join(path, key + ".csv")
                df_convergence = pd.read_csv(import_path, index_col=0, header=None)
                df_convergence.index.rename("iter", inplace=True)
                df_convergence.plot(ax=ax[y_position], legend=False)
            else:
                for source_dim in range(model.source_dimension):
                    # TODO: better legend?
                    import_path = os.path.join(
                        path, key + "_" + str(source_dim) + ".csv"
                    )
                    df_convergence = pd.read_csv(import_path, index_col=0, header=None)
                    df_convergence.index.rename("iter", inplace=True)
                    df_convergence.plot(ax=ax[y_position], legend=False)

        quartiles_factor = 0.6745  # = scipy.stats.norm.ppf(.75)

        for i, key in enumerate(reals_ind_name):
            if skip_sources and key in ["sources"]:
                continue

            import_path_mean = os.path.join(path, f"{key}_mean.csv")
            df_convergence_mean = pd.read_csv(
                import_path_mean, index_col=0, header=None
            )
            df_convergence_mean.index.rename("iter", inplace=True)

            import_path_std = os.path.join(path, f"{key}_std.csv")
            df_convergence_std = pd.read_csv(import_path_std, index_col=0, header=None)
            df_convergence_std.index.rename("iter", inplace=True)

            df_convergence_mean.columns = [f"{key}_mean"]
            df_convergence_std.columns = [f"{key}_std"]  # is it variance or std-dev??

            df_convergence = pd.concat(
                [df_convergence_mean, df_convergence_std], axis=1
            )

            y_position += 1
            df_convergence.plot(
                use_index=True, y=f"{key}_mean", ax=ax[y_position], legend=False
            )

            mu, sd = df_convergence[f"{key}_mean"], df_convergence[f"{key}_std"]
            ax[y_position].fill_between(
                df_convergence.index,
                mu - quartiles_factor * sd,
                mu + quartiles_factor * sd,
                color="b",
                alpha=0.2,
            )
            ax[y_position].set_title(key)

        plt.grid(True)
        plt.tight_layout()
        plt.savefig(path_saveplot_2)
        plt.close()



import json
from typing import Union

from leaspy import __version__
from leaspy.exceptions import LeaspyModelInputError
from leaspy.utils.typing import KwargsType

__all__ = ["ModelSettings"]


class ModelSettings:
    """
    Used in models.base.load to create a model class object from a `json` file.

    Attributes
    ----------
    name : :obj:`str`
        The model's name.

    parameters : :obj:`KwargsType`
        The model's parameters.

    hyperparameters : :obj:`KwargsType`
        The model's hyperparameters.

    Parameters
    ----------
    path_to_model_settings_or_dict : :obj:`dict` or :obj:`str`
        * If a str: path to a json file containing model settings
        * If a dict: content of model settings

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        If the provided settings are not valid or if the file cannot be read.
    """

    def __init__(self, path_to_model_settings_or_dict: Union[str, dict]):
        if isinstance(path_to_model_settings_or_dict, dict):
            settings = path_to_model_settings_or_dict
        elif isinstance(path_to_model_settings_or_dict, str):
            with open(path_to_model_settings_or_dict) as fp:
                settings = json.load(fp)
        else:
            raise LeaspyModelInputError(
                "Bad type for model settings: should be a dict or a path "
                f"as a string, not {type(path_to_model_settings_or_dict)}"
            )

        ModelSettings._check_settings(settings)
        self.name: str = settings["name"].lower()
        self.parameters: KwargsType = settings["parameters"]
        self.hyperparameters: KwargsType = {
            k.lower(): v
            for k, v in settings.items()
            if k not in ("name", "parameters", "hyperparameters", "leaspy_version")
        }

    @staticmethod
    def _check_settings(settings: dict) -> None:
        """Check if the settings are valid.

        Parameters
        ----------
        settings : :obj:`dict`
            The settings to check.

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If the settings are not valid: Key is missing or version is incompatible.
        """
        for mandatory_key in ("name", "parameters"):
            if mandatory_key not in settings:
                raise LeaspyModelInputError(
                    f"The '{mandatory_key}' key is missing in the model "
                    "parameters (JSON file) you are loading."
                )
        if "leaspy_version" not in settings:
            raise LeaspyModelInputError(
                "The model you are trying to load was generated with a leaspy version < 1.1 "
                f"and is not compatible with your current version of leaspy == {__version__} "
                "because of a bug in the multivariate model which lead to under-optimal results.\n"
                "Please consider re-calibrating your model with your current leaspy version.\n"
                "If you really want to load it as is (at your own risk) please use leaspy == 1.0.*"
            )



from abc import ABC, abstractmethod

import torch

from leaspy.exceptions import LeaspyModelInputError
from leaspy.utils.typing import DictParamsTorch, ParamType

__all__ = ["AbstractAttributes"]


class AbstractAttributes(ABC):
    """
    Abstract base class for attributes of models.

    Contains the common attributes & methods of the different attributes classes.
    Such classes are used to update the models' attributes.

    Parameters
    ----------
    name : str
    dimension : int (default None)
    source_dimension : int (default None)

    Attributes
    ----------
    name : str
        Name of the associated leaspy model.
    dimension : int
        Number of features of the model
    source_dimension : int
        Number of sources of the model
        TODO? move to AbstractManifoldModelAttributes?
    univariate : bool
        Whether model is univariate or not (i.e. dimension == 1)
    has_sources : bool
        Whether model has sources or not (not univariate and source_dimension >= 1)
        TODO? move to AbstractManifoldModelAttributes?
    update_possibilities : set[str] (default empty)
        Contains the available parameters to update. Different models have different parameters.

    Raises
    ------
    :exc:`.LeaspyModelInputError`
        if any inconsistent parameter.
    """

    def __init__(self, name: str, dimension: int, source_dimension: int):
        if not (isinstance(name, str) and len(name)):
            raise LeaspyModelInputError(
                "In model attributes, you must provide a non-empty string for the parameter `name`."
            )
        self.name = name

        if not (isinstance(dimension, int) and dimension >= 1):
            raise LeaspyModelInputError(
                "In model attributes, you must provide an integer >= 1 for the parameter `dimension`."
            )
        self.dimension = dimension
        self.univariate = dimension == 1

        self.source_dimension = source_dimension
        self.has_sources = bool(source_dimension)

        if self.univariate and self.has_sources:
            raise LeaspyModelInputError(
                "Inconsistent attributes: presence of sources for a univariate model."
            )

        if not (
            isinstance(source_dimension, int) and (0 <= source_dimension < dimension)
        ):
            raise LeaspyModelInputError(
                "In model attributes, you must provide an integer in [0, dimension - 1] for the parameter `source_dimension`."
            )

        self.update_possibilities: set[ParamType] = set()

    @abstractmethod
    def get_attributes(
        self,
    ) -> tuple[torch.FloatTensor, torch.FloatTensor, torch.FloatTensor]:
        """
        Returns the attributes of the model, which is a tuple of three torch tensors.

        For the precise definitions of those attributes please refer to the exact
        attributes class associated to your model.

        Returns
        -------
        positions: `torch.Tensor`
        velocities: `torch.Tensor`
        mixing_matrix: `torch.Tensor`
        """

    @abstractmethod
    def update(
        self, names_of_changed_values: set[ParamType], values: DictParamsTorch
    ) -> None:
        """
        Update model group average parameter(s).

        Parameters
        ----------
        names_of_changed_values : set[str]
           Values to be updated
        values : dict [str, `torch.Tensor`]
           New values used to update the model's group average parameters

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If `names_of_changed_values` contains unknown values to update.
        """

    def move_to_device(self, device: torch.device):
        """
        Move the tensor attributes of this class to the specified device.

        Parameters
        ----------
        device : torch.device
        """
        for attribute_name in dir(self):
            if attribute_name.startswith("__"):
                continue
            attribute = getattr(self, attribute_name)
            if isinstance(attribute, torch.Tensor):
                setattr(self, attribute_name, attribute.to(device))

    def _check_names(self, names_of_changed_values: set[ParamType]):
        """
        Check if the name of the parameter(s) to update are in the possibilities allowed by the model.

        Parameters
        ----------
        names_of_changed_values : set[str]

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If `names_of_changed_values` contains unknown values to update.
        """
        unknown_update_possibilities = names_of_changed_values.difference(
            self.update_possibilities
        )
        if len(unknown_update_possibilities):
            raise LeaspyModelInputError(
                f"{unknown_update_possibilities} not in the attributes that can be updated"
            )



from typing import Optional

import numpy as np
import pandas as pd

from leaspy.exceptions import LeaspyDataInputError

from .abstract_dataframe_data_reader import AbstractDataframeDataReader
from .individual_data import IndividualData

__all__ = ["VisitDataframeDataReader"]


class VisitDataframeDataReader(AbstractDataframeDataReader):
    """
    Methods to convert :class:`pandas.DataFrame` to `Leaspy`-compliant data containers for longitudinal data only.
    Raises
    ------
    :exc:`.LeaspyDataInputError`
    """

    def __init__(self):
        super().__init__()

    @property
    def dimension(self) -> Optional[int]:
        """

        Number of longitudinal outcomes in dataset.

        Returns
        -------
        : :obj:`int`
            Number of longitudinal outcomes in dataset
        """
        if self.long_outcome_names is None:
            return None
        return len(self.long_outcome_names)

    @classmethod
    def _check_TIME(cls, s: pd.Series) -> None:
        """
        Check requirements on patient's visits indexing: only numeric value and no missing values are tolerated

        Parameters
        ----------
        s: pd.Series
            Pandas series that contains the time at visits of each patient

        """
        if not cls._check_numeric_type(s):
            raise LeaspyDataInputError(
                f"The `TIME` column should contain numeric values (not {s.dtype})."
            )

        s.replace([np.inf, -np.inf], np.nan, inplace=True)
        if s.isna().any():
            individuals_with_at_least_1_bad_tpt = s.isna().groupby("ID").any()
            individuals_with_at_least_1_bad_tpt = individuals_with_at_least_1_bad_tpt[
                individuals_with_at_least_1_bad_tpt
            ].index.tolist()
            raise LeaspyDataInputError(
                "The `TIME` column should NOT contain any nan nor inf, "
                f"please double check these individuals:\n{individuals_with_at_least_1_bad_tpt}."
            )

    def _check_headers(self, columns: list[str]) -> None:
        """
        Check mendatory dataframe headers

        Parameters
        ----------
        columns: List[str]
            Names of the columns headers of the dataframe that contains patients information
        """
        missing_mandatory_columns = [_ for _ in ["ID", "TIME"] if _ not in columns]
        if len(missing_mandatory_columns) > 0:
            raise LeaspyDataInputError(
                f"Your dataframe must have {missing_mandatory_columns} columns"
            )

    def _set_index(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Set the index suited for the type of information contained in the dataframe

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information

        Returns
        -------
        df: pd.DataFrame
            Dataframe with the right index
        """

        # Check and clean visit times
        self._check_TIME(df.set_index("ID")["TIME"])
        df["TIME"] = round(
            df["TIME"], self.time_rounding_digits
        )  # avoid missing duplicates due to rounding errors

        # Set index and make sure it is unique
        df.set_index(["ID", "TIME"], inplace=True)

        return df

    def _clean_dataframe(
        self, df: pd.DataFrame, *, drop_full_nan: bool, warn_empty_column: bool
    ) -> pd.DataFrame:
        """
        Clean the dataframe that contains repeated measures information for each visit

        Parameters
        ----------
        df: pd.DataFrame
            Dataframe with patient information.

        drop_full_nan: bool
            If set to True, raw full of nan are dropped.

        warn_empty_column: bool
            If set to True, a warning is raise for columns full of nan.


        Returns
        -------
        df: pd.DataFrame
            Dataframe with clean information.

        Raises
        ------
        :exc:`.LeaspyDataInputError` :
            - If the :df:`pd.DataFrame` is empty in terms of raw
            - If the :df:`pd.DataFrame` is empty in terms of columns
        """
        self.n_visits = len(df)
        if self.n_visits == 0:
            raise LeaspyDataInputError(
                "Dataframe should have at least 1 row (not full of nans)..."
            )

        self.long_outcome_names = df.columns.tolist()
        if self.dimension < 1:
            raise LeaspyDataInputError("Dataframe should have at least 1 feature...")

        return df

    def _load_individuals_data(
        self, subj: IndividualData, df_subj: pd.DataFrame
    ) -> None:
        """
        Convert information stored in a dataframe to information stored into IndividualData

        Parameters
        ----------
        subj: IndividualData
            One patient with her/his information, potentially empty

        df_subj: pd.DataFrame
            One patient with her/his information
        """
        subj.add_observations(
            timepoints=df_subj.index.get_level_values("TIME").to_list(),
            observations=df_subj[self.long_outcome_names].values.tolist(),
        )



import inspect
import re
import warnings
from functools import partial, reduce
from typing import Callable, Dict, Iterable, Optional, TypeVar

# from typing_extensions import Literal

T = TypeVar("T")
R = TypeVar("R")

DEFAULT_INIT_DOC = "Initialize self.  See help(type(self)) for accurate signature."


def _replace_terms(source: str, mapping: Dict[str, str], flags: int = 0) -> str:
    """
    Replace all occurrences of keys in a string by their mapped correspondence.

    <!> The correspondences are searched with word boundaries and is case-sensitive

    Parameters
    ----------
    source : str
        Source string to replace from
    mapping : dict
        Mapping of terms to replace {original: replacement}
        <!> No replacement term should by an original key to replace
    flags : int
        Valid flag for :func:`re.sub`

    Returns
    -------
    str
        The input str with all replacements done.

    Examples
    --------
    >>> _replace_terms("Can't say the word `less` since I'm wordless. word-", \
                       {'say': 'hear', 'word': '***', 'less': '!?', "I'm": "you're"})
    "Can't hear the *** `!?` since you're wordless. ***-"
    """

    assert (
        len(set(mapping.values()).intersection(mapping.keys())) == 0
    ), "Replacements and replaced should be disjoint."

    return reduce(
        lambda s, repl: re.sub(rf"\b{re.escape(repl[0])}\b", repl[1], s, flags=flags),
        mapping.items(),
        source,
    )


def doc_with_(
    target: object, original: object, mapping: Dict[str, str] = None, **mapping_kwargs
) -> object:
    """
    Document (in-place) a function/class.

    Low-level version of :func:`.doc_with` (refer to its documentation)
    Will set `target.__doc__` in-place (not a decorator function).

    Parameters
    ----------
    target : object
        Object to document (e.g. a function, a class, a given class method, ...).
    original : object
        Object to copy documentation from.
    mapping : dict
        Optional mapping to replace some terms (case-sensitive and word boundary aware) by others
        from the original docstring.
    **mapping_kwargs
        Optional keyword arguments passed to :func:`._replace_terms` (flags=...).

    Returns
    -------
    target
        The (in-place) modified target object.
    """
    original_doc = original.__doc__
    assert original_doc is not None

    if mapping is not None:
        original_doc = _replace_terms(original_doc, mapping, **mapping_kwargs)

    if hasattr(target, "__func__"):  # special method (wrapped) [i.e. classmethod]
        target.__func__.__doc__ = (
            original_doc  # in-place modification of wrapped func doc
        )
    else:
        target.__doc__ = original_doc  # in-place modification

    # we have to return the function for `doc_with` wrapper to be a valid decorator
    return target


def doc_with(
    original: object, mapping: Dict[str, str] = None, **mapping_kwargs
) -> Callable[[object], object]:
    """
    Factory of function/class decorator to use the docstring of `original`
    to document (in-place) the decorated function/class

    Parameters
    ----------
    original : documented Python object
        The object to extract the docstring from
    mapping : dict[str, str], optional
        Optional mapping to replace some terms (case-sensitive and word boundary aware) by others
        from the original docstring.
    **mapping_kwargs
        Optional keyword arguments passed to :func:`._replace_terms` (flags=...).

    Returns
    -------
    Function/class decorator
    """
    return partial(doc_with_, original=original, mapping=mapping, **mapping_kwargs)


def _get_first_candidate(
    candidates: Iterable[T], getter: Callable[[T], Optional[R]]
) -> Optional[R]:
    for c in candidates:
        obj = getter(c)
        if obj is not None:
            return obj
    return None


def _get_attr_if_cond(
    attr_name: str, cond: Optional[Callable[[R], bool]] = None
) -> Callable[[object], Optional[R]]:
    def getter(obj: object) -> Optional[R]:
        attr: R = getattr(obj, attr_name, None)
        return (
            attr if attr is None or cond is None or cond(attr) else None
        )  # lazy bool eval

    return getter


def _get_function_parameters_without_type_annotations(f: Callable):
    # we remove type hints of the parameters to check a loose equality between function input signature
    # because often we will add type hints in super method but not in subclass methods (boring...)
    s = inspect.signature(f)
    params = [
        p.replace(annotation=inspect.Signature.empty) for p in s.parameters.values()
    ]
    # we also ignore positional only flag (essential from magic methods such __str__)
    params = [
        p.replace(kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
        if p.kind is inspect.Parameter.POSITIONAL_ONLY
        else p
        for p in params
    ]
    return params


# def doc_with_super(*, if_other_signature: Literal['force', 'warn', 'skip', 'raise'] = 'force', **doc_with_kwargs) -> Callable[[T], T]:
def doc_with_super(
    *, if_other_signature: str = "raise", **doc_with_kwargs
) -> Callable[[T], T]:
    """
    Factory of class decorator that comment (in-place) all of its inherited methods without docstrings + its top docstring
    with the ones from its parent class (the first parent class with this method documented if multiple inheritance)

    Parameters
    ----------
    if_other_signature : str
        Behavior if a documented method was found in parent but it has another signature:
            * ``'force'``: patch the method with the found docstring anyway (default)
            * ``'warn'``: patch the method but with a warning regarding signature mismatch
            * ``'skip'``: don't patch the method with the found docstring
            * ``'raise'``: raise a ValueError
    **doc_with_kwargs
        Optional keyword arguments passed to :func:`.doc_with` (mapping=...).

    Returns
    -------
    Class decorator
    """

    # what methods are we looking to patch with parent doc (including builtin)
    is_method_without_doc = (
        lambda cls_member: inspect.isroutine(cls_member) and cls_member.__doc__ is None
    )

    # simple condition
    member_has_doc = lambda member: member.__doc__ is not None

    # check doc & signature of candidates methods
    def condition_on_super_method_gen(m: Callable) -> Callable[[Callable], bool]:
        # info on subclass method
        m_name = m.__qualname__
        m_sign = _get_function_parameters_without_type_annotations(m)

        def condition_on_super_method(super_m: Callable) -> bool:
            # ignore not documented methods or default documented __init__ method
            if super_m.__doc__ is None or (
                m_name.endswith(".__init__") and super_m.__doc__ == DEFAULT_INIT_DOC
            ):
                return False

            super_sign = _get_function_parameters_without_type_annotations(super_m)
            sign_is_same = super_sign == m_sign
            if not sign_is_same:
                if if_other_signature == "warn":
                    warnings.warn(
                        f"{m_name} has a different signature than its parent {super_m.__qualname__}, patching doc anyway."
                    )
                    return True
                elif if_other_signature == "raise":
                    raise ValueError(
                        f"{m_name} has a different signature than its parent {super_m.__qualname__}, aborting."
                    )

            # when if_other_signature == 'skip'
            return sign_is_same

        return condition_on_super_method

    # class decorator
    def wrapper(cls):
        assert (
            len(cls.__bases__) > 0
        ), "Must be applied on a class inheriting from others..."

        # patch the class doc itself
        if cls.__doc__ is None:
            super_cls = _get_first_candidate(
                cls.__bases__, lambda kls: kls if kls.__doc__ is not None else None
            )
            if super_cls is not None:
                doc_with_(cls, super_cls, **doc_with_kwargs)  # in-place

        # get all relevant methods to patch and loop on them: list[(m_name:str, m:method)]
        methods = inspect.getmembers(cls, is_method_without_doc)
        for m_name, m in methods:
            # condition on the super method to be a valid candidate to document method `m`
            if if_other_signature == "force":
                cond_on_super_method = member_has_doc
            else:
                cond_on_super_method = condition_on_super_method_gen(m)

            super_method = _get_first_candidate(
                cls.__bases__, _get_attr_if_cond(m_name, cond_on_super_method)
            )

            if super_method is not None:
                doc_with_(m, super_method, **doc_with_kwargs)  # in-place

        return cls

    return wrapper



import torch

from .abstract_manifold_model_attributes import AbstractManifoldModelAttributes

__all__ = ["LogisticAttributes"]


class LogisticAttributes(AbstractManifoldModelAttributes):
    """
    Attributes of leaspy logistic models.

    Contains the common attributes & methods to update the logistic model's attributes.

    Parameters
    ----------
    name : str
    dimension : int
    source_dimension : int

    Attributes
    ----------
    name : str (default 'logistic')
        Name of the associated leaspy model.
    dimension : int
    source_dimension : int
    univariate : bool
        Whether model is univariate or not (i.e. dimension == 1)
    has_sources : bool
        Whether model has sources or not (not univariate and source_dimension >= 1)
    update_possibilities : set[str] (default {'all', 'g', 'v0', 'betas'} )
        Contains the available parameters to update. Different models have different parameters.
    positions : :class:`torch.Tensor` [dimension] (default None)
        positions = exp(realizations['g']) such that "p0" = 1 / (1 + positions)
    velocities : :class:`torch.Tensor` [dimension] (default None)
        Always positive: exp(realizations['v0'])
    orthonormal_basis : :class:`torch.Tensor` [dimension, dimension - 1] (default None)
    betas : :class:`torch.Tensor` [dimension - 1, source_dimension] (default None)
    mixing_matrix : :class:`torch.Tensor` [dimension, source_dimension] (default None)
        Matrix A such that w_i = A * s_i.

    See Also
    --------
    :class:`~leaspy.models.univariate_model.UnivariateModel`
    :class:`~leaspy.models.multivariate_model.MultivariateModel`
    """

    def __init__(self, name, dimension, source_dimension):
        super().__init__(name, dimension, source_dimension)

    def update(self, names_of_changed_values, values):
        """
        Update model group average parameter(s).

        Parameters
        ----------
        names_of_changed_values : set[str]
            Elements of set must be either:
                * ``all`` (update everything)
                * ``g`` correspond to the attribute :attr:`positions`.
                * ``v0`` (only for multivariate models) correspond to the attribute :attr:`velocities`.
                  When we are sure that the v0 change is only a scalar multiplication
                  (in particular, when we reparametrize log(v0) <- log(v0) + mean(xi)),
                  we may update velocities using ``v0_collinear``, otherwise
                  we always assume v0 is NOT collinear to previous value
                  (no need to perform the verification it is - would not be really efficient)
                * ``betas`` correspond to the linear combination of columns from the orthonormal basis so
                  to derive the :attr:`mixing_matrix`.
        values : dict [str, `torch.Tensor`]
            New values used to update the model's group average parameters

        Raises
        ------
        :exc:`.LeaspyModelInputError`
            If `names_of_changed_values` contains unknown parameters.
        """
        self._check_names(names_of_changed_values)

        compute_betas = False
        compute_positions = False
        compute_velocities = False
        dgamma_t0_not_collinear_to_previous = False

        if "all" in names_of_changed_values:
            # make all possible updates
            names_of_changed_values = self.update_possibilities

        if "betas" in names_of_changed_values:
            compute_betas = True
        if "g" in names_of_changed_values:
            compute_positions = True
        if ("v0" in names_of_changed_values) or (
            "v0_collinear" in names_of_changed_values
        ):
            compute_velocities = True
            dgamma_t0_not_collinear_to_previous = "v0" in names_of_changed_values

        if compute_positions:
            self._compute_positions(values)
        if compute_velocities:
            self._compute_velocities(values)

        # only for models with sources beyond this point
        if not self.has_sources:
            return

        if compute_betas:
            self._compute_betas(values)

        # do not recompute orthonormal basis when we know dgamma_t0 is collinear
        # to previous velocities to avoid useless computations!
        recompute_ortho_basis = compute_positions or dgamma_t0_not_collinear_to_previous

        if recompute_ortho_basis:
            self._compute_orthonormal_basis()
        if recompute_ortho_basis or compute_betas:
            self._compute_mixing_matrix()

    def _compute_positions(self, values):
        """
        Update the attribute ``positions``.

        Parameters
        ----------
        values : dict [str, `torch.Tensor`]
        """
        self.positions = torch.exp(values["g"])

    def _compute_orthonormal_basis(self):
        """
        Compute the attribute ``orthonormal_basis`` which is an orthonormal basis, w.r.t the canonical inner product,
        of the sub-space orthogonal, w.r.t the inner product implied by the metric, to the time-derivative of the geodesic at initial time.
        """
        # Compute the diagonal of metric matrix (cf. `_compute_Q`)
        G_metric = (1 + self.positions).pow(4) / self.positions.pow(
            2
        )  # = "1/(p0 * (1-p0))**2"

        dgamma_t0 = self.velocities

        # Householder decomposition in non-Euclidean case, updates `orthonormal_basis` in-place
        self._compute_Q(dgamma_t0, G_metric)



from typing import Optional

import numpy as np
import statsmodels.api as sm
import torch

from leaspy.io.outputs import IndividualParameters
from leaspy.utils.typing import DictParamsTorch

from .stateless import StatelessModel

__all__ = ["LMEModel"]


class LMEModel(StatelessModel):
    r"""LMEModel is a benchmark model that fits and personalize a linear mixed-effects model.

    The model specification is the following:

    .. math:: y_{ij} = fixed_{intercept} + random_{intercept_i} + (fixed_{slopeAge} + random_{slopeAge_i}) * age_{ij} + \epsilon_{ij}

    with:
        * :math:`y_{ij}`: value of the feature of the i-th subject at his j-th visit,
        * :math:`age_{ij}`: age of the i-th subject at his j-th visit.
        * :math:`\epsilon_{ij}`: residual Gaussian noise (independent between visits)

    .. warning::
        This model must be fitted on one feature only (univariate model).

    TODO? add some covariates in this very simple model.

    Parameters
    ----------
    name : :obj:`str`
        The model's name.
    **kwargs
        Model hyperparameters:
            * with_random_slope_age : :obj:`bool` (default ``True``).

    Attributes
    ----------
    name : :obj:`str`
        The model's name.
    is_initialized : :obj:`bool`
        ``True`` if the model is initialized, ``False`` otherwise.
    with_random_slope_age : :obj:`bool` (default ``True``)
        Has the LME a random slope for subject's age?
        Otherwise it only has a random intercept per subject.
    features : :obj:`list` of :obj:`str`
        List of the model features.

        .. warning::
            LME has only one feature.

    dimension : :obj:`int`
        Will always be 1 (univariate).
    parameters : :obj:`dict`
        Contains the model parameters. In particular:
            * ``ages_mean`` : :obj:`float`
                Mean of ages (for normalization).
            * ``ages_std`` : :obj:`float`
                Std-dev of ages (for normalization).
            * ``fe_params`` : :class:`np.ndarray` of :obj:`float`
                Fixed effects.
            * ``cov_re`` : :class:`np.ndarray`
                Variance-covariance matrix of random-effects.
            * ``cov_re_unscaled_inv`` : :class:`np.ndarray`
                Inverse of unscaled (= divided by variance of noise) variance-covariance matrix of random-effects.
                This matrix is used for personalization to new subjects.
            * ``noise_std`` : :obj:`float`
                Std-dev of Gaussian noise.
            * ``bse_fe``, ``bse_re`` : :class:`np.ndarray` of :obj:`float`
                Standard errors on fixed-effects and random-effects respectively (not used in ``Leaspy``).

    See Also
    --------
    :class:`~leaspy.algo.others.lme_fit.LMEFitAlgorithm`
    :class:`~leaspy.algo.others.lme_personalize.LMEPersonalizeAlgorithm`
    """

    def __init__(
        self, name: str, with_random_slope_age: Optional[bool] = True, **kwargs
    ):
        super().__init__(name, **kwargs)
        self.with_random_slope_age = with_random_slope_age
        self.dimension = 1

    @property
    def hyperparameters(self) -> DictParamsTorch:
        """Dictionary of values for model hyperparameters."""
        return {}

    def compute_individual_trajectory(
        self,
        timepoints: list[float],
        individual_parameters: IndividualParameters,
    ) -> torch.Tensor:
        """Compute scores values at the given time-point(s) given a subject's individual parameters.

        Parameters
        ----------
        timepoints : array-like of ages (not normalized)
            Timepoints to compute individual trajectory at.

        individual_parameters : :obj:`dict`
            Individual parameters:
                * random_intercept
                * random_slope_age (if ``with_random_slope_age == True``)

        Returns
        -------
        :class:`torch.Tensor` of :obj:`float` :
            The individual trajectories. The shape of the tensor is
            ``(n_individuals == 1, n_tpts == len(timepoints), n_features == 1)``.
        """
        # normalize ages (np.ndarray of float, 1D)
        ages_norm = (
            np.array(timepoints).reshape(-1) - self.parameters["ages_mean"]
        ) / self.parameters["ages_std"]

        # design matrix (same for fixed and random effects)
        X = sm.add_constant(ages_norm, prepend=True, has_constant="add")

        # assert 'random_intercept' in individual_parameters
        if not self.with_random_slope_age:
            # no random slope on ages (fixed effect only)
            re_params = np.array([individual_parameters["random_intercept"].item(), 0])
        else:
            # assert 'random_slope_age' in individual_parameters
            re_params = np.array(
                [
                    individual_parameters["random_intercept"].item(),
                    individual_parameters["random_slope_age"].item(),
                ]
            )
        y = X @ (self.parameters["fe_params"] + re_params)

        return torch.tensor(y, dtype=torch.float32).reshape((1, -1, 1))



import torch

from leaspy.io.data.dataset import Dataset
from leaspy.utils.functional import Exp, OrthoBasis
from leaspy.utils.weighted_tensor import (
    TensorOrWeightedTensor,
    WeightedTensor,
    unsqueeze_right,
)
from leaspy.variables.distributions import Normal
from leaspy.variables.specs import (
    Hyperparameter,
    LinkedVariable,
    ModelParameter,
    NamedVariables,
    PopulationLatentVariable,
    VariableNameToValueMapping,
)

from .logistic import LogisticInitializationMixin
from .time_reparametrized import TimeReparametrizedModel

__all__ = ["SharedSpeedLogisticModel"]


class SharedSpeedLogisticModel(LogisticInitializationMixin, TimeReparametrizedModel):
    """
    Logistic model for multiple variables of interest, imposing same average
    evolution pace for all variables (logistic curves are only time-shifted).

    Parameters
    ----------
    name : :obj:`str`
        The name of the model.
    **kwargs
        Hyperparameters of the model.
    """

    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)

    def _compute_initial_values_for_model_parameters(
        self,
        dataset: Dataset,
    ) -> VariableNameToValueMapping:
        parameters = super()._compute_initial_values_for_model_parameters(dataset)
        parameters["log_g_mean"] = parameters["log_g_mean"].mean()
        parameters["xi_mean"] = parameters["log_v0_mean"].mean()
        del parameters["log_v0_mean"]
        parameters["deltas_mean"] = torch.zeros((self.dimension - 1,))
        return parameters

    @staticmethod
    def metric(*, g_deltas_exp: torch.Tensor) -> torch.Tensor:
        """
        Compute the metric term for the logistic model.

        This scaling term modulates the curvature of the logistic trajectory.

        Parameters
        ----------
        g_deltas_exp : :class:`torch.Tensor`
          Product of slope and exp(-deltas).

        Returns
        -------
        :class:`torch.Tensor`
         Metric value, computed as:
               .. math::

                    \\frac{(g \\cdot e^{-\\delta} + 1)^2}{g \\cdot e^{-\\delta}}
        """
        return (g_deltas_exp + 1) ** 2 / g_deltas_exp

    @staticmethod
    def deltas_exp(*, deltas_padded: torch.Tensor) -> torch.Tensor:
        """Compute the exponential of the negative deltas.

        Parameters:
            deltas_padded (:class:`torch.Tensor`): Padded deltas.

        Returns:
            :class:`torch.Tensor`: Exponential of the negative deltas.
        """
        return torch.exp(-1 * deltas_padded)

    @staticmethod
    def g_deltas_exp(*, g: torch.Tensor, deltas_exp: torch.Tensor) -> torch.Tensor:
        return g * deltas_exp

    @staticmethod
    def pad_deltas(*, deltas: torch.Tensor) -> torch.Tensor:
        """Prepend deltas with a zero as delta_1 is set to zero in the equations.
        .
                Parameters:
                    deltas (:class:`torch.Tensor`): Deltas tensor.

                Returns:
                    :class:`torch.Tensor`: Padded deltas tensor.
        """
        return torch.cat((torch.tensor([0.0]), deltas))

    @staticmethod
    def denom(*, g_deltas_exp: torch.Tensor) -> torch.Tensor:
        r"""
        Compute the denominator for the :math:`\gamma_{t_0}` calculation.
        """
        return 1 + g_deltas_exp

    @staticmethod
    def gamma_t0(*, denom: torch.Tensor) -> torch.Tensor:
        r"""
        Compute the :math:`\gamma_{t_0}` value, which is the inverse of the denominator.

        Parameters
        ----------
        denom : :class:`torch.Tensor`
            Denominator term.

        Returns
        -------
        :class:`torch.Tensor`
            :math:`\gamma_{t_0}` value.
        """
        return 1 / denom

    @staticmethod
    def g_metric(*, gamma_t0: torch.Tensor) -> torch.Tensor:
        """
        Compute the g_metric value, which is the square of gamma_t0.

        Parameters
        ----------
        gamma_t0 : :class:`torch.Tensor`
            :math:`\\gamma_{t_0}` value.

        Returns
        -------
        :class:`torch.Tensor`
            :math:`g_{metric}` value, computed as:
               .. math::

                    g\\_metric = \\frac{1}{(\\gamma_{t0} \\cdot (1 - \\gamma_{t0}))^2}
        """
        return 1 / (gamma_t0 * (1 - gamma_t0)) ** 2

    @staticmethod
    def collin_to_d_gamma_t0(
        *, deltas_exp: torch.Tensor, denom: torch.Tensor
    ) -> torch.Tensor:
        """Compute the collinear term to d_gamma_t0.

        Parameters
        ----------
        deltas_exp : :class:`torch.Tensor`
            Exponential of the negative deltas.
        denom : :class:`torch.Tensor`
            Denominator term.

        Returns
        -------
        :class:`torch.Tensor`
            Collinear term to d_gamma_t0
        """
        return deltas_exp / denom**2

    @classmethod
    def model_with_sources(
        cls,
        *,
        rt: TensorOrWeightedTensor[float],
        space_shifts: TensorOrWeightedTensor[float],
        metric: TensorOrWeightedTensor[float],
        deltas_padded: TensorOrWeightedTensor[float],
        log_g: TensorOrWeightedTensor[float],
    ) -> torch.Tensor:
        """Returns a model with sources."""
        # Shape: (Ni, Nt, Nfts)
        pop_s = (None, None, ...)
        rt = unsqueeze_right(rt, ndim=1)  # .filled(float('nan'))
        w_model_logit = (
            metric[pop_s] * space_shifts[:, None, ...]
            + rt
            + deltas_padded
            - log_g[pop_s]
        )
        model_logit, weights = WeightedTensor.get_filled_value_and_weight(
            w_model_logit, fill_value=0.0
        )
        return WeightedTensor(torch.sigmoid(model_logit), weights).weighted_value

    @classmethod
    def model_no_sources(
        cls,
        *,
        rt: TensorOrWeightedTensor[float],
        metric: TensorOrWeightedTensor[float],
        deltas_padded: TensorOrWeightedTensor[float],
        log_g: TensorOrWeightedTensor[float],
    ) -> torch.Tensor:
        """Returns a model without sources."""
        return cls.model_with_sources(
            rt=rt,
            metric=metric,
            deltas_padded=deltas_padded,
            log_g=log_g,
            space_shifts=torch.zeros((1, 1)),
        )

    def get_variables_specs(self) -> NamedVariables:
        """
        Get the specifications of the variables used in the model.
        """
        d = super().get_variables_specs()
        d.update(
            log_g_mean=ModelParameter.for_pop_mean("log_g", shape=(1,)),
            log_g_std=Hyperparameter(0.01),
            log_g=PopulationLatentVariable(Normal("log_g_mean", "log_g_std")),
            g=LinkedVariable(Exp("log_g")),
            xi_mean=ModelParameter.for_ind_mean("xi", shape=(1,)),
            deltas_mean=ModelParameter.for_pop_mean(
                "deltas",
                shape=(self.dimension - 1,),
            ),
            deltas_std=Hyperparameter(0.01),
            deltas=PopulationLatentVariable(
                Normal("deltas_mean", "deltas_std"),
                sampling_kws={"scale": 1.0},
            ),
            deltas_padded=LinkedVariable(self.pad_deltas),
            deltas_exp=LinkedVariable(self.deltas_exp),
            g_deltas_exp=LinkedVariable(self.g_deltas_exp),
            metric=LinkedVariable(self.metric),
        )
        if self.source_dimension >= 1:
            d.update(
                denom=LinkedVariable(self.denom),
                gamma_t0=LinkedVariable(self.gamma_t0),
                collin_to_d_gamma_t0=LinkedVariable(self.collin_to_d_gamma_t0),
                g_metric=LinkedVariable(self.g_metric),
                orthonormal_basis=LinkedVariable(
                    OrthoBasis("collin_to_d_gamma_t0", "g_metric"),
                ),
                model=LinkedVariable(self.model_with_sources),
            )
        else:
            d["model"] = LinkedVariable(self.model_no_sources)

        return d



from bisect import bisect
from typing import Any, Optional

import numpy as np
import pandas as pd

from leaspy.exceptions import LeaspyDataInputError, LeaspyInputError, LeaspyTypeError
from leaspy.utils.typing import FeatureType, IDType

__all__ = ["IndividualData"]


class IndividualData:
    """
    Container for an individual's data

    Parameters
    ----------
    idx : IDType
        Unique ID

    Attributes
    ----------
    idx : :class:`~leaspy.utils.typing.IDType`
        Unique ID
    timepoints : :obj:`np.ndarray` [:obj:`float`]
        Timepoints associated with the observations 1D array
    observations : :obj:`np.ndarray` [:obj:`float`]
        Observed data points,
        Shape is ``(n_timepoints, n_features)``
    cofactors : :obj:`dict` [:class:`~leaspy.utils.typing.FeatureType`, :class:`~leaspy.utils.typing.Any`]
        Cofactors in the form {cofactor_name: cofactor_value}
    event_time : :obj:`float`
        Time of an event, if the event is censored, the time correspond to the last patient observation
    event_bool : :obj:`bool`
        Boolean to indicate if an event is censored or not: 1 observed, 0 censored
    """

    def __init__(self, idx: IDType):
        self.idx: IDType = idx
        self.timepoints: np.ndarray = None
        self.observations: np.ndarray = None
        self.event_time: Optional[np.ndarray] = None
        self.event_bool: Optional[np.ndarray] = None
        self.cofactors: dict[FeatureType, Any] = {}
        self.covariates: Optional[np.ndarray] = None

    def add_observations(
        self, timepoints: list[float], observations: list[list[float]]
    ) -> None:
        """
        Include new observations and associated timepoints

        Parameters
        ----------
        timepoints : :obj:`array-like` [:obj:`float`]
            Timepoints associated with the observations to include, 1D array
        observations : :obj:`array-like` [:obj:`float`]
            Observations to include, 2D array

        Raises
        ------
        :exc:`.LeaspyDataInputError`
        """
        for t, obs in zip(timepoints, observations):
            if self.timepoints is None:
                self.timepoints = np.array([t])
                self.observations = np.array([obs])
            elif t in self.timepoints:
                raise LeaspyDataInputError(
                    f"Trying to overwrite timepoint {t} " f"of individual {self.idx}"
                )
            else:
                index = bisect(self.timepoints, t)
                self.timepoints = np.concatenate(
                    [self.timepoints[:index], [t], self.timepoints[index:]]
                )
                self.observations = np.concatenate(
                    [self.observations[:index], [obs], self.observations[index:]]
                )

    def add_event(self, event_time: list[float], event_bool: list[bool]) -> None:
        """
        Include event time and associated censoring bool

        Parameters
        ----------
        event_time : :obj:`float`
            Time of the event
        event_bool : :obj:`float`
            0 if censored (not observed) and 1 if observed

        """
        self.event_time = np.array(event_time)
        self.event_bool = np.array(event_bool)

    def add_covariates(self, covariates: list[list[int]]) -> None:
        """
        Include covariates

        Parameters
        ----------
        covariates : :obj:`array-like` [:obj:`float`]
            Covariates to include, 2D array
        """
        self.covariates = np.array(covariates)

    def add_cofactors(self, cofactors: dict[FeatureType, Any]) -> None:
        """
        Include new cofactors

        Parameters
        ----------
        cofactors : :obj:`dict` [:class:`~leaspy.utils.typing.FeatureType`, :class:`~leaspy.utils.typing.Any`]
            Cofactors to include, in the form `{name: value}`

        Raises
        ------
        :exc:`.LeaspyDataInputError`
        :exc:`.LeaspyTypeError`
        """
        if not (
            isinstance(cofactors, dict)
            and all(
                isinstance(cofactor_name, str) for cofactor_name in cofactors.keys()
            )
        ):
            raise LeaspyTypeError("Invalid argument type for `cofactors`")

        for cofactor_name, cofactor_value in cofactors.items():
            if (
                cofactor_name in self.cofactors
                and cofactor_value != self.cofactors[cofactor_name]
            ):
                raise LeaspyDataInputError(
                    f"Cofactor {cofactor_name} is already present for patient {self.idx} "
                    f"with a value of {self.cofactors[cofactor_name]} different from the value "
                    f"{cofactor_value} that you are trying to set."
                )
            self.cofactors[cofactor_name] = cofactor_value

    def to_frame(
        self,
        headers: list,
        event_time_name: str,
        event_bool_name: str,
        covariate_names: list[str],
    ) -> pd.DataFrame:
        """
        Convert the individual data to a pandas DataFrame

        Parameters
        ----------
        headers : :obj:`list` [:obj:`str`]
            List of feature names for the observations
        event_time_name : :obj:`str`
            Name of the column for the event time
        event_bool_name : :obj:`str`
            Name of the column for the event boolean (0 or 1)
        covariate_names : :obj:`list` [:obj:`str`]
            List of covariate names

        Returns
        -------
        :obj:`pd.DataFrame`
            DataFrame containing the individual's data with the following columns:
                * ID: Unique identifier for the individual
                * TIME: Timepoints associated with the observations
                * Observations: Observed data points for each feature
                * Event Time: Time of the event (if any)
                * Event Boolean: Boolean indicating if the event was observed (1) or censored (0)
                * Covariates: Values of the covariates for the individual

        """
        type_to_concat = []
        if self.observations is not None:
            ix_tpts = pd.MultiIndex.from_product(
                [[self.idx], self.timepoints], names=["ID", "TIME"]
            )
            type_to_concat.append(
                pd.DataFrame(self.observations, columns=headers, index=ix_tpts)
            )
        if self.event_time is not None:
            df_event = self._event_to_frame(event_time_name, event_bool_name)
            type_to_concat.append(df_event)

        if self.covariates is not None:
            df_covariate = self._covariate_to_frame(covariate_names)
            type_to_concat.append(df_covariate)

        if len(type_to_concat) == 1:
            return type_to_concat[0]
        else:
            return type_to_concat[1].join(type_to_concat[0])

    def _event_to_frame(
        self, event_time_name: str, event_bool_name: str
    ) -> pd.DataFrame:
        """
        Convert the event data to a pandas DataFrame

        Parameters
        ----------
        event_time_name : :obj:`str`
            Name of the column for the event time
        event_bool_name : :obj:`str`
            Name of the column for the event boolean (0 or 1)

        Returns
        -------
        :obj:`pd.DataFrame`
            DataFrame containing the event data with the following columns:
            - ID: Unique identifier for the individual
            - Event Time: Time of the event (if any)
            - Event Boolean: Boolean indicating if the event was observed (1) or censored (0)
        """
        ix_tpts = pd.Index([self.idx], name="ID")
        if len(np.unique(self.event_time)) != 1:
            raise LeaspyInputError(
                f"Individual {self.idx} has multiple time at event only one is accepted"
            )

        if self.event_bool.sum() == 1:
            event_coded = np.where(self.event_bool)[0][0]
            event_bool = event_coded + 1
        elif self.event_bool.sum() == 0:
            event_bool = 0
        else:
            raise LeaspyInputError(
                f"Individual {self.idx} should contain maximum one observed event"
            )

        df_event = pd.DataFrame(
            data=[[self.event_time[0], event_bool]],
            index=ix_tpts,
            columns=[event_time_name, event_bool_name],
        )
        df_event[event_time_name] = df_event[event_time_name].astype(float)
        df_event[event_bool_name] = df_event[event_bool_name].astype(int)
        return df_event

    def _covariate_to_frame(self, covariate_names: list[str]) -> pd.DataFrame:
        """
        Convert the covariates to a pandas DataFrame

        Parameters
        ----------
        covariate_names : :obj:`list`[:obj:`str`]
            List of covariate names

        Returns
        -------
        :obj:`pd.DataFrame`
            DataFrame containing the covariates with the following columns:
            - ID: Unique identifier for the individual
            - Covariates: Values of the covariates for the individual
        """
        ix_tpts = pd.Index([self.idx], name="ID")
        df_covariates = pd.DataFrame(
            data=[self.covariates],
            index=ix_tpts,
            columns=[covariate_names],
        )
        return df_covariates



from functools import wraps
from typing import Callable, Optional, TypeVar

import torch

from ._weighted_tensor import TensorOrWeightedTensor, WeightedTensor

__all__ = [
    "factory_weighted_tensor_unary_operator",
]


VT = TypeVar("VT")


def factory_weighted_tensor_unary_operator(
    f: Callable[[torch.Tensor], torch.Tensor],
    *,
    fill_value: Optional[VT] = None,
) -> Callable[[TensorOrWeightedTensor[VT]], TensorOrWeightedTensor[VT]]:
    """Factory/decorator to create a weighted-tensor compatible function from the provided unary-tensor function."""

    @wraps(f)
    def f_compatible(
        x: TensorOrWeightedTensor[VT], *args, **kws
    ) -> TensorOrWeightedTensor[VT]:
        if not isinstance(x, WeightedTensor):
            return f(x, *args, **kws)
        r = f(x.filled(fill_value), *args, **kws)
        conv = x.valued
        if isinstance(r, (tuple, list, set, frozenset)):
            return type(r)(map(conv, r))
        return conv(r)

    return f_compatible



import warnings

import numpy as np

from leaspy.exceptions import LeaspyAlgoInputError

from .settings import AlgorithmSettings

__all__ = ["AlgorithmWithAnnealingMixin"]


class AlgorithmWithAnnealingMixin:
    """Mixin class to use in algorithms that requires `temperature_inv`.

    Note that this mixin should be used with a class inheriting from `AbstractAlgo`, which must have `algo_parameters`
    attribute.

    Parameters
    ----------
    settings : :class:`.AlgorithmSettings`
        The specifications of the algorithm as a :class:`.AlgorithmSettings` instance.

        Please note that you can customize the number of iterations with annealing by setting:
             * `annealing.n_iter_frac`, such that iterations with annealing is a ratio of algorithm `n_iter` (default = 50%)

    Attributes
    ----------
    annealing_on : :obj:`bool`
        Activates annealing.

    temperature : float >= 1
    temperature_inv : float in [0, 1]
        Temperature and its inverse when using annealing
    """

    def __init__(self, settings: AlgorithmSettings):
        super().__init__(settings)
        self.temperature: float = 1.0
        self.temperature_inv: float = 1.0
        # useful property derived from algo parameters
        self.annealing_on: bool = self.algo_parameters.get("annealing", {}).get(
            "do_annealing", False
        )
        self._annealing_period: int = None
        self._annealing_temperature_decrement: float = None

        if not self.annealing_on:
            return

        # Dynamic number of iterations for annealing
        annealing_n_iter_frac = self.algo_parameters["annealing"]["n_iter_frac"]

        if self.algo_parameters["annealing"].get("n_iter", None) is None:
            if annealing_n_iter_frac is None:
                raise LeaspyAlgoInputError(
                    "You should NOT have both `annealing.n_iter_frac` and `annealing.n_iter` None."
                    "\nPlease set a value for at least one of those settings."
                )

            self.algo_parameters["annealing"]["n_iter"] = int(
                annealing_n_iter_frac * self.algo_parameters["n_iter"]
            )

        elif annealing_n_iter_frac is not None:
            warnings.warn(
                "`annealing.n_iter` setting is deprecated in favour of `annealing.n_iter_frac` - "
                "which defines the duration with annealing as a ratio of the total number of iterations."
                "\nPlease use the new setting to suppress this warning "
                "or explicitly set `annealing.n_iter_frac=None`."
                "\nHowever, note that while `annealing.n_iter` is supported "
                "it will always have priority over `annealing.n_iter_frac`.",
                FutureWarning,
            )

    def __str__(self):
        out = super().__str__()
        if self.annealing_on:
            out += "\n= Annealing =\n"
            out += f"    temperature : {self.temperature:.1f}"
        return out

    def _initialize_annealing(self):
        """
        Initialize annealing, setting initial temperature and number of iterations.
        """
        if not self.annealing_on:
            return

        self.temperature = self.algo_parameters["annealing"]["initial_temperature"]
        self.temperature_inv = 1 / self.temperature

        if not (
            isinstance(self.algo_parameters["annealing"]["n_plateau"], int)
            and self.algo_parameters["annealing"]["n_plateau"] > 0
        ):
            raise LeaspyAlgoInputError(
                "Your `annealing.n_plateau` should be a positive integer"
            )

        if self.algo_parameters["annealing"]["n_plateau"] == 1:
            warnings.warn(
                "You defined `annealing.n_plateau` = 1, so you will stay at initial temperature. "
                "Consider setting `annealing.n_plateau` >= 2 for a true annealing scheme."
            )
            return

        self._annealing_period = self.algo_parameters["annealing"]["n_iter"] // (
            self.algo_parameters["annealing"]["n_plateau"] - 1
        )

        self._annealing_temperature_decrement = (
            self.algo_parameters["annealing"]["initial_temperature"] - 1.0
        ) / (self.algo_parameters["annealing"]["n_plateau"] - 1)
        if self._annealing_temperature_decrement <= 0:
            raise LeaspyAlgoInputError("Your `initial_temperature` should be > 1")

    def _update_temperature(self):
        """
        Update the temperature according to a plateau annealing scheme.
        """
        if not self.annealing_on or self._annealing_period is None:
            return

        if self.current_iteration <= self.algo_parameters["annealing"]["n_iter"]:
            # If we cross a plateau step
            if self.current_iteration % self._annealing_period == 0:
                # Oscillating scheme
                if self.algo_parameters["annealing"].get("oscillations", False):
                    params = self.algo_parameters["annealing"]
                    b = params["range"]
                    c = params["delay"]
                    r = params["period"]
                    k = self.current_iteration
                    kappa = c + 2.0 * float(k) * np.pi / r
                    self.temperature = max(1.0 + b * np.sin(kappa) / kappa, 0.1)

                else:
                    # Decrease temperature linearly
                    self.temperature -= self._annealing_temperature_decrement
                    self.temperature = max(self.temperature, 1)

                self.temperature_inv = 1.0 / self.temperature



from .base import FitAlgorithm
from .fit_output_manager import FitOutputManager
from .lme_fit import LMEFitAlgorithm
from .mcmc_saem import TensorMcmcSaemAlgorithm

__all__ = [
    "FitAlgorithm",
    "TensorMcmcSaemAlgorithm",
    "FitOutputManager",
    "LMEFitAlgorithm",
]
